# ruff: noqa: F401
import os
import re
from collections import Counter, defaultdict
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import spacy.tokenizer
from fsspec import filesystem as fsspec
from loguru import logger

from edsnlp import registry
from edsnlp.core import PipelineProtocol
from edsnlp.core.lazy_collection import LazyCollection
from edsnlp.data.base import BaseReader, BaseWriter
from edsnlp.data.converters import (
    FILENAME,
    AttributesMappingArg,
    SequenceStr,
    get_dict2doc_converter,
    get_doc2dict_converter,
)
from edsnlp.utils.collections import flatten_once
from edsnlp.utils.file_system import FileSystem, normalize_fs_path, walk_match
from edsnlp.utils.span_getters import SpanSetterArg

REGEX_ENTITY = re.compile(r"^(T\d+)\t(.*) (\d+ \d+(?:;\d+ \d+)*)\t(.*)$")
REGEX_NOTE = re.compile(r"^(#\d+)\tAnnotatorNotes ([^\t]+)\t(.*)$")
REGEX_RELATION = re.compile(r"^(R\d+)\t(\S+) Arg1:(\S+) Arg2:(\S+)")
REGEX_ATTRIBUTE = re.compile(r"^([AM]\d+)\t(.+?) ([TE]\d+)(?: (.+))?$")
REGEX_EVENT = re.compile(r"^(E\d+)\t(.+)$")
REGEX_EVENT_PART = re.compile(r"(\S+):([TE]\d+)")


class BratParsingError(ValueError):
    def __init__(self, ann_file, line):
        super().__init__(f"File {ann_file}, unrecognized Brat line {line}")


LOCAL_FS = fsspec("file")


def parse_standoff_file(
    path: str,
    merge_spaced_fragments: bool = True,
    fs: FileSystem = LOCAL_FS,
) -> Dict:
    """
    Load a brat file

    Adapted from
    https://github.com/percevalw/nlstruct/blob/master/nlstruct/datasets/brat.py

    Parameters
    ----------
    path: str
        Path or glob path of the brat text file (.txt, not .ann)
    merge_spaced_fragments: bool
        Merge fragments of an entity that was split by brat because it overlapped an
        end of line
    fs: FileSystem
        Filesystem to use

    Returns
    -------
    Iterator[Dict]
    """
    ann_filenames = []
    for filename in walk_match(
        fs,
        os.path.dirname(path),
        os.path.basename(path).replace(".txt", ".a*"),
        recursive=False,
    ):
        ann_filenames.append(filename)

    entities = {}
    relations = []
    events = {}

    with fs.open(path, "r") as f:
        text = f.read()

    if not len(ann_filenames):
        return {
            "text": text,
        }

    for ann_file in ann_filenames:
        with fs.open(ann_file, "r") as f:
            for line_idx, line in enumerate(f):
                try:
                    if line.startswith("T"):
                        match = REGEX_ENTITY.match(line)
                        if match is None:
                            raise BratParsingError(ann_file, line)
                        ann_id = match.group(1)
                        entity = match.group(2)
                        span = match.group(3)
                        mention_text = match.group(4)
                        entities[ann_id] = {
                            "text": mention_text,
                            "entity_id": ann_id,
                            "fragments": [],
                            "attributes": {},
                            "notes": [],
                            "label": entity,
                        }
                        last_end = None
                        fragment_i = 0
                        begins_ends = sorted(
                            [
                                (int(s.split()[0]), int(s.split()[1]))
                                for s in span.split(";")
                            ]
                        )

                        for begin, end in begins_ends:
                            # If merge_spaced_fragments, merge two fragments that are
                            # only separated by a newline (brat automatically creates
                            # multiple fragments for a entity that spans over more than
                            # one line)
                            if (
                                merge_spaced_fragments
                                and last_end is not None
                                and len(text[last_end:begin].strip()) == 0
                            ):
                                entities[ann_id]["fragments"][-1]["end"] = end
                                last_end = end
                                continue
                            entities[ann_id]["fragments"].append(
                                {
                                    "begin": begin,
                                    "end": end,
                                }
                            )
                            fragment_i += 1
                            last_end = end
                    elif line.startswith("A") or line.startswith("M"):
                        match = REGEX_ATTRIBUTE.match(line)
                        if match is None:
                            raise BratParsingError(ann_file, line)
                        _, attr_name, entity_id, value = match.groups()
                        if attr_name is None:
                            raise BratParsingError(ann_file, line)
                        (
                            entities[entity_id]
                            if entity_id.startswith("T")
                            else events[entity_id]
                        )["attributes"][attr_name] = value
                    elif line.startswith("R"):
                        match = REGEX_RELATION.match(line)
                        if match is None:
                            raise BratParsingError(ann_file, line)
                        ann_id = match.group(1)
                        ann_name = match.group(2)
                        arg1 = match.group(3)
                        arg2 = match.group(4)
                        relations.append(
                            {
                                "relation_id": ann_id,
                                "relation_label": ann_name,
                                "from_entity_id": arg1,
                                "to_entity_id": arg2,
                            }
                        )
                    elif line.startswith("E"):
                        match = REGEX_EVENT.match(line)
                        if match is None:
                            raise BratParsingError(ann_file, line)
                        ann_id = match.group(1)
                        arguments_txt = match.group(2)
                        arguments = []
                        for argument in REGEX_EVENT_PART.finditer(arguments_txt):
                            arguments.append(
                                {
                                    "entity_id": argument.group(2),
                                    "label": argument.group(1),
                                }
                            )
                        events[ann_id] = {
                            "event_id": ann_id,
                            "attributes": {},
                            "arguments": arguments,
                        }
                    elif line.startswith("#"):
                        match = REGEX_NOTE.match(line)
                        if match is None:
                            raise BratParsingError(ann_file, line)
                        ann_id = match.group(1)
                        entity_id = match.group(2)
                        note = match.group(3)
                        entities[entity_id]["notes"].append(
                            {
                                "note_id": ann_id,
                                "value": note,
                            }
                        )
                except Exception:
                    raise Exception(
                        "Could not parse line {} from {}: {}".format(
                            line_idx, ann_file, repr(line)
                        )
                    )
    return {
        "text": text,
        "entities": list(entities.values()),
        "relations": relations,
        "events": list(events.values()),
    }


def dump_standoff_file(
    doc,
    txt_filename,
    overwrite_txt=False,
    overwrite_ann=False,
    fs: FileSystem = LOCAL_FS,
):
    parent_dir = txt_filename.rsplit("/", 1)[0]
    if parent_dir and not fs.exists(parent_dir):
        fs.makedirs(parent_dir, exist_ok=True)
    if not fs.exists(txt_filename) or overwrite_txt:
        with fs.open(txt_filename, "w") as f:
            f.write(doc["text"])

    ann_filename = txt_filename.replace(".txt", ".ann")
    attribute_idx = 1
    entities_ids = defaultdict(lambda: "T" + str(len(entities_ids) + 1))
    if not fs.exists(ann_filename) or overwrite_ann:
        with fs.open(ann_filename, "w") as f:
            if "entities" in doc:
                for entity in doc["entities"]:
                    spans = []
                    brat_entity_id = entities_ids[entity["entity_id"]]
                    for fragment in sorted(
                        entity["fragments"], key=lambda frag: frag["begin"]
                    ):
                        idx = fragment["begin"]
                        entity_text = doc["text"][fragment["begin"] : fragment["end"]]
                        # eg: "mon entité  \n  est problématique"
                        for match in re.finditer(
                            r"\s*(.+?)(?:( *\n+)+ *|$)", entity_text, flags=re.DOTALL
                        ):
                            spans.append((idx + match.start(1), idx + match.end(1)))
                    print(
                        "{}\t{} {}\t{}".format(
                            brat_entity_id,
                            str(entity["label"]),
                            ";".join(" ".join(map(str, span)) for span in spans),
                            " ".join(doc["text"][begin:end] for begin, end in spans),
                        ),
                        file=f,
                    )
                    if "attributes" in entity:
                        for i, (key, value) in enumerate(entity["attributes"].items()):
                            if value:
                                print(
                                    "A{}\t{} {}{}".format(
                                        attribute_idx,
                                        str(key),
                                        brat_entity_id,
                                        (" " + str(value)) if value is not True else "",
                                    ),
                                    file=f,
                                )
                                attribute_idx += 1

                    # fmt: off
                    # if "relations" in doc:
                    #     for i, relation in enumerate(doc["relations"]):
                    #         entity_from = entities_ids[relation["from_entity_id"]]
                    #         entity_to = entities_ids[relation["to_entity_id"]]
                    #         print(
                    #             "R{}\t{} Arg1:{} Arg2:{}\t".format(
                    #                 i + 1, str(relation["label"]), entity_from,
                    #                 entity_to
                    #             ),
                    #             file=f,
                    #         )
                    # fmt: on


class StandoffReader(BaseReader):
    DATA_FIELDS = ()

    def __init__(
        self,
        path: Union[str, Path],
        *,
        keep_ipynb_checkpoints: bool = False,
        keep_txt_only_docs: bool = False,
        filesystem: Optional[FileSystem] = None,
    ):
        super().__init__()
        self.fs, self.path = normalize_fs_path(filesystem, path)
        self.files: List[str] = [
            file
            for file in walk_match(self.fs, self.path, "*.txt")
            if (keep_ipynb_checkpoints or ".ipynb_checkpoints" not in str(file))
            and (
                keep_txt_only_docs
                or walk_match(
                    self.fs,
                    os.path.dirname(file),
                    os.path.basename(file).replace(".txt", ".a*"),
                    recursive=False,
                )
            )
        ]
        assert len(self.files), f"No .txt files found in the BRAT directory {self.path}"
        for file in self.files:
            if not self.fs.exists(file):
                raise FileNotFoundError(f"File {file} does not exist")
        logger.info(f"The BRAT directory contains {len(self.files)} .txt files.")

    def read_main(self) -> Iterable[Tuple[str, int]]:
        return ((f, 1) for f in self.files)

    def read_worker(self, fragment: List[str]):
        tasks = []
        for file in fragment:
            anns = parse_standoff_file(str(file), fs=self.fs)
            anns[FILENAME] = os.path.relpath(file, self.path).rsplit(".", 1)[0]
            anns["doc_id"] = anns[FILENAME]
            tasks.append(anns)
        return tasks


class StandoffWriter(BaseWriter):
    def __init__(
        self,
        path: Union[str, Path],
        *,
        lines: bool = True,
        overwrite: bool = False,
        filesystem: Optional[FileSystem] = None,
    ):
        self.fs, self.path = normalize_fs_path(filesystem, path)
        self.lines = lines

        if self.fs.exists(self.path):
            unsafe_exts = Counter(
                os.path.splitext(f)[1] for f in walk_match(self.fs, self.path, "*.txt")
            ) + Counter(
                os.path.splitext(f)[1] for f in walk_match(self.fs, self.path, "*.a*")
            )
            if unsafe_exts and not overwrite:
                raise FileExistsError(
                    f"Directory {self.path} already exists and appear to contain "
                    "annotations:"
                    + "".join(f"\n -{s[1:]}: {v} files" for s, v in unsafe_exts.items())
                    + "\nUse overwrite=True to write files anyway."
                )
        self.fs.makedirs(self.path, exist_ok=True)

        super().__init__()

    def write_worker(self, records: List[Dict]) -> Tuple[List[str], int]:
        results = []
        for rec in records:
            filename = str(rec[FILENAME])
            path = os.path.join(self.path, f"{filename}.txt")
            dump_standoff_file(
                rec,
                path,
                overwrite_txt=True,
                overwrite_ann=True,
                fs=self.fs,
            )
            results.append(path)
        return results, len(results)

    def write_main(self, fragments) -> List[str]:
        return list(flatten_once(fragments))


# noinspection PyIncorrectDocstring
@registry.readers.register("standoff")
def read_standoff(
    path: Union[str, Path],
    *,
    keep_ipynb_checkpoints: bool = False,
    keep_txt_only_docs: bool = False,
    converter: Optional[Union[str, Callable]] = "standoff",
    filesystem: Optional[FileSystem] = None,
    **kwargs,
) -> LazyCollection:
    """
    The BratReader (or `edsnlp.data.read_standoff`) reads a directory of BRAT files and
    yields documents. At the moment, only entities and attributes are loaded. Relations
     and events are not supported.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.read_standoff("path/to/brat/directory")
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.read_standoff` returns a
        [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list :

        ```{ .python .no-check }
        docs = list(edsnlp.data.read_standoff("path/to/brat/directory"))
        ```

    !!! warning "True/False attributes"

        Boolean values are not supported by the BRAT editor, and are stored as empty
        (key: empty value) if true, and not stored otherwise. This means that False
        values will not be assigned to attributes by default, which can be problematic
        when deciding if an entity is negated or not : is the entity not negated, or
        has the negation attribute not been annotated ?

        To avoid this issue, you can use the `bool_attributes` argument to specify
        which attributes should be considered as boolean when reading a BRAT dataset.
        These attributes will be assigned a value of `True` if they are present, and
        `False` otherwise.

        ```{ .python .no-check }
        doc_iterator = edsnlp.data.read_standoff(
            "path/to/brat/directory",
            # Mapping from 'BRAT attribute name' to 'Doc attribute name'
            span_attributes={"Negation": "negated"},
            bool_attributes=["negated"],  # Missing values will be set to False
        )
        ```

    Parameters
    ----------
    path : Union[str, Path]
        Path to the directory containing the BRAT files (will recursively look for
        files in subdirectories).
    nlp : Optional[PipelineProtocol]
        The pipeline object (optional and likely not needed, prefer to use the
        `tokenizer` directly argument instead).
    tokenizer : Optional[spacy.tokenizer.Tokenizer]
        The tokenizer instance used to tokenize the documents. Likely not needed since
        by default it uses the current context tokenizer :

        - the tokenizer of the next pipeline run by `.map_pipeline` in a
          [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        - or the `eds` tokenizer by default.
    span_setter : SpanSetterArg
        The span setter to use when setting the spans in the documents. Defaults to
        setting the spans in the `ents` attribute, and creates a new span group for
        each JSON entity label.
    span_attributes : Optional[AttributesMappingArg]
        Mapping from BRAT attributes to Span extensions (can be a list too).
        By default, all attributes are imported as Span extensions with the same name.
    keep_raw_attribute_values : bool
        Whether to keep the raw attribute values (as strings) or to convert them to
        Python objects (e.g. booleans).
    default_attributes : AttributesMappingArg
        How to set attributes on spans for which no attribute value was found in the
        input format. This is especially useful for negation, or frequent attributes
        values (e.g. "negated" is often False, "temporal" is often "present"), that
        annotators may not want to annotate every time.
    notes_as_span_attribute : Optional[str]
        If set, the AnnotatorNote annotations will be concatenated and stored in a span
        attribute with this name.
    split_fragments : bool
        Whether to split the fragments into separate spans or not. If set to False, the
        fragments will be concatenated into a single span.
    keep_ipynb_checkpoints : bool
        Whether to keep the files that are in the `.ipynb_checkpoints` directory.
    keep_txt_only_docs : bool
        Whether to keep the `.txt` files that do not have corresponding `.ann` files.
    converter : Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects.
    filesystem: Optional[FileSystem] = None,
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).

    Returns
    -------
    LazyCollection
    """
    data = LazyCollection(
        reader=StandoffReader(
            path,
            keep_ipynb_checkpoints=keep_ipynb_checkpoints,
            keep_txt_only_docs=keep_txt_only_docs,
            filesystem=filesystem,
        )
    )
    if converter:
        converter, kwargs = get_dict2doc_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)
    return data


@registry.writers.register("standoff")
def write_standoff(
    data: Union[Any, LazyCollection],
    path: Union[str, Path],
    overwrite: bool = False,
    converter: Optional[Union[str, Callable]] = "standoff",
    filesystem: Optional[FileSystem] = None,
    **kwargs,
) -> None:
    """
    `edsnlp.data.write_standoff` writes a list of documents using the BRAT/Standoff
    format in a directory. The BRAT files will be named after the `note_id` attribute of
    the documents, and subdirectories will be created if the name contains `/`
    characters.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)

    doc = nlp("My document with entities")

    edsnlp.data.write_standoff([doc], "path/to/brat/directory")
    ```

    !!! warning "Overwriting files"

        By default, `write_standoff` will raise an error if the directory already exists
        and contains files with `.a*` or `.txt` suffixes. This is to avoid overwriting
        existing annotations. To allow overwriting existing files, use `overwrite=True`.

    Parameters
    ----------
    data: Union[Any, LazyCollection],
        The data to write (either a list of documents or a LazyCollection).
    path: Union[str, Path]
        Path to the directory containing the BRAT files (will recursively look for
        files in subdirectories).
    span_getter: SpanGetterArg
        The span getter to use when listing the spans that will be exported as BRAT
        entities. Defaults to getting the spans in the `ents` attribute.
    span_attributes: Optional[AttributesMappingArg]
        Mapping from BRAT attributes to Span extension. By default, no attribute will
        be exported.
    overwrite: bool
        Whether to overwrite existing directories.
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects.
        Defaults to the "standoff" format converter.
    filesystem: Optional[FileSystem] = None,
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).
    """
    data = LazyCollection.ensure_lazy(data)
    if converter:
        converter, kwargs = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)

    return data.write(StandoffWriter(path, overwrite=overwrite, filesystem=filesystem))



from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

from spacy import Errors, Vocab, registry, util
from spacy.language import DEFAULT_CONFIG, FactoryMeta, Language
from spacy.pipe_analysis import validate_attrs
from spacy.tokens import Doc
from spacy.util import SimpleFrozenDict, SimpleFrozenList, raise_error
from spacy.vocab import create_vocab

from edsnlp.core.registries import accepted_arguments


@classmethod
def factory(
    cls,
    name: str,
    *,
    default_config: Dict[str, Any] = SimpleFrozenDict(),
    assigns: Iterable[str] = SimpleFrozenList(),
    requires: Iterable[str] = SimpleFrozenList(),
    retokenizes: bool = False,
    default_score_weights: Dict[str, Optional[float]] = SimpleFrozenDict(),
    func: Optional[Callable] = None,
) -> Callable:
    """
    Patched from spaCy to allow back dots in factory
    names (https://github.com/aphp/edsnlp/pull/152)

    Register a new pipeline component factory. Can be used as a decorator
    on a function or classmethod, or called as a function with the factory
    provided as the func keyword argument. To create a component and add
    it to the pipeline, you can use nlp.add_pipe(name).

    name (str): The name of the component factory.
    default_config (Dict[str, Any]): Default configuration, describing the
        default values of the factory arguments.
    assigns (Iterable[str]): Doc/Token attributes assigned by this component,
        e.g. "token.ent_id". Used for pipeline analysis.
    requires (Iterable[str]): Doc/Token attributes required by this component,
        e.g. "token.ent_id". Used for pipeline analysis.
    retokenizes (bool): Whether the component changes the tokenization.
        Used for pipeline analysis.
    default_score_weights (Dict[str, Optional[float]]): The scores to report during
        training, and their default weight towards the final score used to
        select the best model. Weights should sum to 1.0 per component and
        will be combined and normalized for the whole pipeline. If None,
        the score won't be shown in the logs or be weighted.
    func (Optional[Callable]): Factory function if not used as a decorator.

    DOCS: https://spacy.io/api/language#factory
    """
    if not isinstance(name, str):
        raise ValueError(Errors.E963.format(decorator="factory"))
    if not isinstance(default_config, dict):
        raise ValueError(
            Errors.E962.format(
                style="default config", name=name, cfg_type=type(default_config)
            )
        )

    def add_factory(factory_func: Callable) -> Callable:
        internal_name = cls.get_factory_name(name)
        if internal_name in registry.factories.namespace:
            # We only check for the internal name here – it's okay if it's a
            # subclass and the base class has a factory of the same name. We
            # also only raise if the function is different to prevent raising
            # if module is reloaded.
            existing_func = registry.factories.get(internal_name)
            if not util.is_same_func(factory_func, existing_func):
                raise ValueError(
                    Errors.E004.format(
                        name=name, func=existing_func, new_func=factory_func
                    )
                )

        util.get_arg_names(factory_func)
        if len(accepted_arguments(factory_func, ["nlp", "name"])) < 2:
            raise ValueError(Errors.E964.format(name=name))
        # Officially register the factory so we can later call
        # registry.resolve and refer to it in the config as
        # @factories = "spacy.Language.xyz". We use the class name here so
        # different classes can have different factories.
        registry.factories.register(internal_name, func=factory_func)
        factory_meta = FactoryMeta(
            factory=name,
            default_config=default_config,
            assigns=validate_attrs(assigns),
            requires=validate_attrs(requires),
            scores=list(default_score_weights.keys()),
            default_score_weights=default_score_weights,
            retokenizes=retokenizes,
        )
        cls.set_factory_meta(name, factory_meta)
        # We're overwriting the class attr with a frozen dict to handle
        # backwards-compat (writing to Language.factories directly). This
        # wouldn't work with an instance property and just produce a
        # confusing error – here we can show a custom error
        registry.factories.entry_points = False
        cls.factories = SimpleFrozenDict(
            registry.factories.get_all(), error=Errors.E957
        )
        registry.factories.entry_points = True
        return factory_func

    if func is not None:  # Support non-decorator use cases
        return add_factory(func)
    return add_factory


def __init__(
    self,
    vocab: Union[Vocab, bool] = True,
    *,
    max_length: int = 10**6,
    meta: Dict[str, Any] = {},
    create_tokenizer: Optional[Callable[["Language"], Callable[[str], Doc]]] = None,
    create_vectors: Optional[Callable[["Vocab"], Any]] = None,
    batch_size: int = 1000,
    **kwargs,
) -> None:  # pragma: no cover
    """
    EDS-NLP: Patched from spaCy do enable lazy-loading components

    Initialise a Language object.

    vocab (Vocab): A `Vocab` object. If `True`, a vocab is created.
    meta (dict): Custom meta data for the Language class. Is written to by
        models to add model meta data.
    max_length (int): Maximum number of characters in a single text. The
        current models may run out memory on extremely long texts, due to
        large internal allocations. You should segment these texts into
        meaningful units, e.g. paragraphs, subsections etc, before passing
        them to spaCy. Default maximum length is 1,000,000 charas (1mb). As
        a rule of thumb, if all pipeline components are enabled, spaCy's
        default models currently requires roughly 1GB of temporary memory per
        100,000 characters in one text.
    create_tokenizer (Callable): Function that takes the nlp object and
        returns a tokenizer.
    batch_size (int): Default batch size for pipe and evaluate.

    DOCS: https://spacy.io/api/language#init
    """

    # EDS-NLP: disable spacy default call to load every factory
    # since some of them may be missing dependencies (like torch)
    util.registry._entry_point_factories.get_all()

    self._config = DEFAULT_CONFIG.merge(self.default_config)
    self._meta = dict(meta)
    self._path = None
    self._optimizer = None
    # Component meta and configs are only needed on the instance
    self._pipe_meta: Dict[str, "FactoryMeta"] = {}  # meta by component
    self._pipe_configs = {}  # config by component

    if not isinstance(vocab, Vocab) and vocab is not True:
        raise ValueError(Errors.E918.format(vocab=vocab, vocab_type=type(Vocab)))
    if vocab is True:
        vectors_name = meta.get("vectors", {}).get("name")
        vocab = create_vocab(self.lang, self.Defaults, vectors_name=vectors_name)
        if (
            not create_vectors
            and "vectors" in self._config["nlp"]
            and "@vectors" in self._config["nlp"]["vectors"]
        ):
            vectors_cfg = {"vectors": self._config["nlp"]["vectors"]}
            create_vectors = registry.resolve(vectors_cfg)["vectors"]
        if create_vectors:
            vocab.vectors = create_vectors(vocab)
    else:
        if (self.lang and vocab.lang) and (self.lang != vocab.lang):
            raise ValueError(Errors.E150.format(nlp=self.lang, vocab=vocab.lang))
    self.vocab: Vocab = vocab
    if self.lang is None:
        self.lang = self.vocab.lang
    self._components: List[Tuple[str, Callable[[Doc], Doc]]] = []
    self._disabled: Set[str] = set()
    self.max_length = max_length
    # Create the default tokenizer from the default config
    if not create_tokenizer:
        tokenizer_cfg = {"tokenizer": self._config["nlp"]["tokenizer"]}
        create_tokenizer = registry.resolve(tokenizer_cfg)["tokenizer"]
    self.tokenizer = create_tokenizer(self)
    self.batch_size = batch_size
    self.default_error_handler = raise_error


Language.factory = factory
Language.__init__ = __init__



from .atomic.days import (  # lz_numeric_day_pattern,
    day_pattern,
    letter_day_pattern,
    numeric_day_pattern,
)
from .atomic.delimiters import (
    ante_num_pattern,
    delimiters,
    post_num_pattern,
    post_num_with_letter_pattern,
    raw_delimiter_with_spaces_pattern,
    raw_delimiters,
)
from .atomic.modes import mode_pattern
from .atomic.months import (
    letter_month_pattern,
    lz_numeric_month_pattern,
    month_pattern,
    numeric_month_pattern,
)
from .atomic.time import time_pattern
from .atomic.years import full_year_pattern as fy_pattern
from .atomic.years import year_pattern

absolute_pattern = []
absolute_pattern_with_time = []

for time in [True, False]:
    pattern = [
        ante_num_pattern
        + day_pattern
        + d
        + month_pattern
        + d
        + year_pattern
        + (time_pattern if time else "")
        + post_num_pattern
        for d in delimiters
    ] + [
        ante_num_pattern
        + year_pattern
        + d
        + numeric_month_pattern
        + d
        + numeric_day_pattern
        + (time_pattern if time else "")
        + post_num_pattern
        for d in delimiters
    ]

    no_year_pattern = [
        day
        + raw_delimiter_with_spaces_pattern
        + month
        + (time_pattern if time else "")
        + post_num_pattern
        for day in [ante_num_pattern + numeric_day_pattern, letter_day_pattern]
        for month in [letter_month_pattern]
    ]
    no_year_pattern.extend(
        [
            ante_num_pattern
            + numeric_day_pattern
            + d
            + numeric_month_pattern
            + post_num_pattern
            for d in raw_delimiters
        ]
    )
    pattern.extend(no_year_pattern)

    no_day_pattern = [
        letter_month_pattern
        + raw_delimiter_with_spaces_pattern
        + year_pattern
        + post_num_with_letter_pattern,
        ante_num_pattern
        + lz_numeric_month_pattern
        + "/"
        + year_pattern
        + post_num_with_letter_pattern,
    ]
    pattern.extend(no_day_pattern)

    no_day_no_year_pattern = [
        letter_month_pattern,
    ]
    pattern.extend(no_day_no_year_pattern)

    full_year_pattern = ante_num_pattern + fy_pattern + post_num_pattern

    pattern.append(full_year_pattern)

    pattern = [r"(?<=" + mode_pattern + r".{,3})?" + p for p in pattern]

    if time:
        absolute_pattern_with_time[:] = pattern
    else:
        absolute_pattern[:] = pattern



from edsnlp.core import registry

from .patterns import sections
from .sections import SectionsMatcher

DEFAULT_CONFIG = dict(
    sections=sections,
    add_patterns=True,
    attr="NORM",
    ignore_excluded=True,
)

create_component = registry.factory.register(
    "eds.sections",
    assigns=["doc.spans", "doc.ents"],
    deprecated=["sections"],
)(SectionsMatcher)



import re

from edsnlp.utils.resources import get_AVC_care_site

from ..terms import BRAIN, HEART, PERIPHERAL

AVC_CARE_SITES_REGEX = [
    r"\b" + re.escape(cs.strip()) + r"\b" for cs in get_AVC_care_site(prefix=True)
] + [
    r"h[oô]p",
    r"\brcp",
    r"service",
    r"\bsau",
    r"ap.?hp",
    r"\burg",
    r"finess",
    r"\bsiret",
    r"[àa] avc",
    r"consult",
]

avc = dict(
    source="avc",
    regex=[
        r"\bavc\b",
    ],
    exclude=[
        dict(
            regex=AVC_CARE_SITES_REGEX,
            window=(-5, 5),
            regex_flags=re.S | re.I,
            limit_to_sentence=False,
        ),
        dict(
            regex=r"\b[a-z]\.",
            window=2,
            limit_to_sentence=False,
        ),
    ],
    regex_attr="NORM",
)

with_localization = dict(
    source="with_localization",
    regex=[
        r"(hemorr?agie|hematome)",
        r"angiopath",
        r"angioplasti",
        r"infarctus",
        r"occlusion",
        r"saignement",
        r"embol",
        r"vascularite",
        r"\bhsd\b",
        r"thrombos",
        r"thrombol[^y]",
        r"thrombophi",
        r"thrombi[^n]",
        r"thrombus",
        r"thrombectomi",
        r"phleb",
    ],
    regex_attr="NORM",
    exclude=[
        dict(
            regex=r"pulmo|poumon",
            window=4,
        ),
    ],
    assign=[
        dict(
            name="brain_localized",
            regex="(" + r"|".join(BRAIN) + ")",
            window=(-15, 15),
            limit_to_sentence=False,
            include_assigned=False,
        ),
    ],
)

general = dict(
    source="general",
    regex=[
        r"accident.{1,5}vasculaires.{1,5}cereb",
        r"accident.{1,5}vasculaire.{1,5}ischemi",
        r"accident.{1,5}ischemi",
        r"moya.?moya",
        r"occlusion.{1,5}(artere|veine).{1,20}retine",
        r"vasculopathies?.cerebrales?.ischemique",
        r"maladies?.des.petites.arteres",
        r"maladies?.des.petits.vaisseaux",
        r"thrombolyse",
        r"\bsusac\b",
    ],
    regex_attr="NORM",
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bAIC\b",
        r"\bOACR\b",
        r"\bOVCR\b",
    ],
    regex_attr="TEXT",
)

AIT = dict(
    source="AIT",
    regex=[
        r"\bAIC\b",
        r"\bOACR\b",
        r"\bOVCR\b",
        r"\bAIT\b",
    ],
    regex_attr="TEXT",
)

ischemia = dict(
    source="ischemia",
    regex=[
        r"ischemi",
    ],
    exclude=[
        dict(
            regex=PERIPHERAL + HEART,
            window=(-7, 7),
        ),
    ],
    assign=[
        dict(
            name="brain",
            regex="(" + r"|".join(BRAIN) + ")",
            window=(-10, 15),
        ),
    ],
    regex_attr="NORM",
)

default_patterns = [
    avc,
    with_localization,
    general,
    acronym,
    AIT,
    ischemia,
]



from edsnlp.core import registry

from .patterns import default_patterns
from .peptic_ulcer_disease import PepticUlcerDiseaseMatcher

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="peptic_ulcer_disease",
    span_setter={"ents": True, "peptic_ulcer_disease": True},
)

create_component = registry.factory.register(
    "eds.peptic_ulcer_disease",
    assigns=["doc.ents", "doc.spans"],
)(PepticUlcerDiseaseMatcher)



results_solid_tumor = dict(
    has_match=[
        True,
        True,
        False,
        True,
        True,
        True,
        True,
    ],
    detailled_status=[
        "LOCALIZED",
        "LOCALIZED",
        None,
        "METASTASIS",
        "METASTASIS",
        "LOCALIZED",
        "METASTASIS",
    ],
    assign=None,
    texts=[
        "Présence d'un carcinome intra-hépatique.",
        "Patient avec un K sein.",
        "Il y a une tumeur bénigne",
        "Tumeur métastasée",
        "Cancer du poumon au stade 4",
        "Cancer du poumon au stade 2",
        "Présence de nombreuses lésions secondaires",
    ],
)



results_liver_disease = dict(
    has_match=4 * [True],
    detailled_status=[
        "MILD",
        "MILD",
        "MODERATE_TO_SEVERE",
        "MODERATE_TO_SEVERE",
    ],
    assign=None,
    texts=[
        "Il y a une fibrose hépatique",
        "Une hépatite B chronique",
        "Le patient consulte pour une cirrhose",
        "Greffe hépatique.",
    ],
)



from typing import Dict, List, Union

ListOrStr = Union[List[str], str]
DictOrPattern = Union[str, List[str], Dict[str, Union[str, List[str]]]]
Patterns = Dict[str, DictOrPattern]


def normalize_token_attr(attr):
    if attr.startswith("doc.") or attr.startswith("span."):
        return None
    attr = attr.replace("token.", "")
    lower = attr.replace("_", "").lower()
    return "text" if lower == "orth" else lower


ATTRIBUTES = {
    "LOWER": "lower_",
    "TEXT": "text",
    "NORM": "norm_",
    "SHAPE": "shape_",
}

from .offset import alignment  # noqa: E402, F401
from .text import get_text  # noqa: E402, F401



import re
import warnings
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol, registry
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.core.contextual_matcher import ContextualMatcher


class SimpleScoreMatcher(ContextualMatcher):
    """
    Matcher component to extract a numeric score

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    label : str
        The name of the extracted score
    regex : List[str]
        A list of regexes to identify the score
    attr : str
        Whether to match on the text ('TEXT') or on the normalized text ('NORM')
    value_extract : str
        Regex with capturing group to get the score value
    score_normalization : Union[str, Callable[[Union[str,None]], Any]]
        Function that takes the "raw" value extracted from the `value_extract`
        regex and should return:

        - None if no score could be extracted
        - The desired score value else
    window : int
        Number of token to include after the score's mention to find the
        score's value
    ignore_excluded : bool
        Whether to ignore excluded spans when matching
    ignore_space_tokens : bool
        Whether to ignore space tokens when matching
    flags : Union[re.RegexFlag, int]
        Regex flags to use when matching
    score_name: str
        Deprecated, use `label` instead. The name of the extracted score
    label : str
        Label name to use for the `Span` object and the extension
    span_setter: Optional[SpanSetterArg]
        How to set matches on the doc
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str,
        *,
        regex: List[str] = None,
        attr: str = "NORM",
        value_extract: Union[str, Dict[str, str], List[Dict[str, str]]] = None,
        score_normalization: Union[str, Callable[[Union[str, None]], Any]] = None,
        window: int = 7,
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        flags: Union[re.RegexFlag, int] = 0,
        score_name: str = None,
        label: str = None,
        span_setter: Optional[SpanSetterArg] = None,
    ):
        if score_name is not None:
            warnings.warn(
                "`score_name` is deprecated, use `label` instead.",
                DeprecationWarning,
            )
            label = score_name

        if label is None:
            raise ValueError("`label` parameter is required.")

        if span_setter is None:
            span_setter = {"ents": True, label: True}

        if isinstance(value_extract, str):
            value_extract = dict(
                name="value",
                regex=value_extract,
                window=window,
            )

        if isinstance(value_extract, dict):
            value_extract = [value_extract]

        value_exists = False
        for i, extract in enumerate(value_extract):
            extract["window"] = extract.get("window", window)
            if extract.get("name", None) == "value":
                value_exists = True
                extract["replace_entity"] = True
                extract["reduce_mode"] = "keep_first"
            value_extract[i] = extract

        assert value_exists, "You should provide a `value` regex in the `assign` dict."

        patterns = dict(
            source=label,
            regex=regex,
            assign=value_extract,
        )

        super().__init__(
            nlp=nlp,
            name=name,
            patterns=patterns,
            assign_as_span=False,
            alignment_mode="expand",
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
            attr=attr,
            regex_flags=flags,
            include_assigned=False,
            label=label,
            span_setter=span_setter,
        )

        if isinstance(score_normalization, str):
            self.score_normalization = registry.misc.get(score_normalization)
        else:
            self.score_normalization = score_normalization

    def set_extensions(self) -> None:
        super().set_extensions()
        if not Span.has_extension(self.label):
            Span.set_extension(self.label, default=None)
        if not Span.has_extension("score_name"):
            Span.set_extension("score_name", default=None)
        if not Span.has_extension("score_value"):
            Span.set_extension("score_value", getter=lambda x: x._.value)

    def process(self, doc: Doc) -> Iterable[Span]:
        """
        Extracts, if available, the value of the score.
        Normalizes the score via the provided `self.score_normalization` method.

        Parameters
        ----------
        doc: Doc
            Document to process

        Yields
        ------
        Span
            Matches with, if found, an added `score_value` extension
        """
        for ent in super().process(doc):
            value = ent._.assigned.get("value", None)
            if value is None:
                continue
            normalized_value = self.score_normalization(value)
            if normalized_value is not None:
                ent._.score_name = self.label
                ent._.set(self.label, normalized_value)

                yield ent



import datetime
from enum import Enum
from typing import Dict, Optional, Union

import pendulum
from pandas._libs.tslibs.nattype import NaTType
from pydantic import BaseModel, root_validator, validator
from spacy.tokens import Span

from edsnlp.pipes.misc.dates.patterns.relative import specific_dict


class Direction(str, Enum):
    FUTURE = "future"
    PAST = "past"
    CURRENT = "current"


class Bound(str, Enum):
    UNTIL = "until"
    FROM = "from"


class Mode(str, Enum):
    ABSOLUTE = "absolute"
    RELATIVE = "relative"
    DURATION = "duration"


class Period(BaseModel):
    FROM: Optional[Span] = None
    UNTIL: Optional[Span] = None
    DURATION: Optional[Span] = None

    class Config:
        arbitrary_types_allowed = True


class BaseDate(BaseModel):
    mode: Mode = None
    bound: Optional[Bound] = None

    @validator("*", pre=True)
    def remove_space(cls, v):
        """Remove spaces. Useful for coping with ill-formatted PDF extractions."""
        if isinstance(v, str):
            return v.replace(" ", "")
        return v


class AbsoluteDate(BaseDate):
    mode: Mode = Mode.ABSOLUTE
    year: Optional[int] = None
    month: Optional[int] = None
    day: Optional[int] = None
    hour: Optional[int] = None
    minute: Optional[int] = None
    second: Optional[int] = None

    def to_datetime(
        self,
        note_datetime: Optional[Union[pendulum.datetime, datetime.datetime]] = None,
        tz: Union[str, pendulum.timezone] = "Europe/Paris",
        infer_from_context: Optional[bool] = None,
        default_day=1,
        default_month=1,
        **kwargs,
    ) -> Optional[pendulum.datetime]:
        """
        Convert the date to a pendulum.datetime object.

        Parameters
        ----------
        tz : Optional[Union[str, pendulum.timezone]]
            The timezone to use. Defaults to "Europe/Paris".
        note_datetime : Optional[Union[pendulum.datetime, datetime.datetime]]
            The datetime of the note. Used to infer missing parts of the date.
        infer_from_context : bool
            Whether to infer missing parts of the date from the note datetime.
            In a (year, month, day) triplet:

                - if only year is missing, it will be inferred from the note datetime
                - if only month is missing, it will be inferred from the note datetime
                - if only day is missing, it will be set to `default_day`
                - if only the year is given, the day and month will be set to
                  `default_day` and `default_month`
                - if only the month is given, the day will be set to `default_day`
                  and the year will be inferred from the note datetime
                - if only the day is given, the month and year will be inferred from
                  the note datetime
        default_day : int
            Default day to use when inferring missing parts of the date.
        default_month : int
            Default month to use when inferring missing parts of the date.

        Returns
        -------
        Optional[pendulum.datetime]
        """

        d = self.dict(exclude_none=True)
        d.pop("mode", None)
        d.pop("bound", None)

        if self.year and self.month and self.day:
            try:
                return pendulum.datetime(**d, tz=tz)
            except ValueError:
                return None
        elif (
            infer_from_context
            or infer_from_context is None
            and note_datetime is not None
        ):
            if note_datetime and not isinstance(note_datetime, NaTType):
                note_datetime = pendulum.instance(note_datetime)

                if self.year is None:
                    d["year"] = note_datetime.year
                if self.month is None:
                    if self.day is None:
                        d["month"] = default_month
                    else:
                        d["month"] = note_datetime.month
                if self.day is None:
                    d["day"] = default_day
            else:
                if self.year is None:
                    return None
                if self.month is None:
                    d["month"] = default_month
                if self.day is None:
                    d["day"] = default_day

            try:
                return pendulum.datetime(**d, tz=tz)
            except ValueError:
                return None

        return None

    def to_duration(
        self,
        note_datetime: Optional[Union[pendulum.datetime, datetime.datetime]] = None,
        **kwargs,
    ) -> Optional[pendulum.Duration]:
        if note_datetime and not isinstance(note_datetime, NaTType):
            note_datetime = pendulum.instance(note_datetime)
            dt = self.to_datetime(note_datetime=note_datetime, **kwargs)
            delta = dt.diff(note_datetime)
            return delta.as_interval()
        else:
            return None

    def norm(self) -> str:
        year = str(self.year) if self.year else "????"
        month = f"{self.month:02}" if self.month else "??"
        day = f"{self.day:02}" if self.day else "??"

        norm = "-".join([year, month, day])

        if self.hour:
            norm += f" {self.hour:02}h"

        if self.minute:
            norm += f"{self.minute:02}m"

        if self.second:
            norm += f"{self.second:02}s"

        return norm

    @validator("year")
    def validate_year(cls, v):
        if v > 100:
            return v

        if v < 25:
            return 2000 + v

    def __str__(self):
        return self.norm()


class Relative(BaseDate):
    mode: Mode = Mode.RELATIVE
    year: Optional[int] = None
    month: Optional[int] = None
    week: Optional[int] = None
    day: Optional[int] = None
    hour: Optional[int] = None
    minute: Optional[int] = None
    second: Optional[int] = None

    @root_validator(pre=True)
    def parse_unit(cls, d: Dict[str, str]) -> Dict[str, str]:
        """
        Units need to be handled separately.

        This validator modifies the key corresponding to the unit
        with the detected value

        Parameters
        ----------
        d : Dict[str, str]
            Original data

        Returns
        -------
        Dict[str, str]
            Transformed data
        """
        unit = d.get("unit")

        if unit:
            d[unit] = d.get("number")

        return d

    def to_duration(self, note_datetime=None, **kwargs) -> pendulum.Duration:
        d = self.dict(exclude_none=True)

        direction = d.pop("direction", None)
        dir = -1 if direction == Direction.PAST else 1

        d.pop("mode", None)

        d = {f"{k}s": v for k, v in d.items()}

        td = dir * pendulum.duration(**d)
        return td

    def to_datetime(self, **kwargs) -> Optional[pendulum.datetime]:
        # for compatibility
        return None


class RelativeDate(Relative):
    direction: Direction = Direction.CURRENT

    def to_datetime(
        self,
        note_datetime: Optional[Union[pendulum.datetime, datetime.datetime]] = None,
        **kwargs,
    ) -> Optional[pendulum.datetime]:
        if note_datetime is not None and not isinstance(note_datetime, NaTType):
            note_datetime = pendulum.instance(note_datetime)

            d = self.dict(exclude_none=True)

            direction = d.pop("direction", None)
            dir = -1 if direction == Direction.PAST else 1

            d.pop("mode", None)
            d.pop("bound", None)

            d = {f"{k}s": v for k, v in d.items()}

            td = dir * pendulum.duration(**d)

            return note_datetime + td

        return None

    def norm(self) -> str:
        if self.direction == Direction.CURRENT:
            d = self.dict(exclude_none=True)
            d.pop("direction", None)
            d.pop("mode", None)

            key = next(iter(d.keys()), "day")

            norm = f"~0 {key}"
        else:
            td = self.to_duration()
            norm = str(td)
            if td.in_seconds() > 0:
                norm = f"+{norm}"

        return norm

    @root_validator(pre=True)
    def handle_specifics(cls, d: Dict[str, str]) -> Dict[str, str]:
        """
        Specific patterns such as `aujourd'hui`, `hier`, etc,
        need to be handled separately.

        Parameters
        ----------
        d : Dict[str, str]
            Original data.

        Returns
        -------
        Dict[str, str]
            Modified data.
        """

        specific = d.get("specific")
        specific = specific_dict.get(specific)

        if specific:
            d.update(specific)

        return d

    def __str__(self):
        return self.norm()


class Duration(Relative):
    mode: Mode = Mode.DURATION

    def norm(self) -> str:
        td = self.to_duration()
        return f"during {td}"

    def to_duration(self, note_datetime=None, **kwargs) -> pendulum.Duration:
        d = self.dict(exclude_none=True)

        d = {f"{k}s": v for k, v in d.items() if k not in ("mode", "bound")}

        return pendulum.duration(**d)



import spacy
from typing import Iterable, List, Optional
import builtins

from libcpp cimport bool as cbool

from spacy.attrs cimport IS_ALPHA, IS_ASCII, IS_DIGIT, IS_LOWER, IS_PUNCT, IS_SPACE
from spacy.lexeme cimport Lexeme
from spacy.tokens.token cimport TokenC

from .terms import punctuation

cdef class FastSentenceSegmenter(object):
    """
    Segments the Doc into sentences using a rule-based strategy,
    specific to AP-HP documents.

    Applies the same rule-based pipeline as spaCy's sentencizer,
    and adds a simple rule on the new lines : if a new line is followed by a
    capitalised word, then it is also an end of sentence.

    DOCS: https://spacy.io/api/sentencizer

    Arguments
    ---------
    punct_chars : Optional[List[str]]
        Punctuation characters.
    use_endlines : bool
        Whether to use endlines prediction.
    """

    def __init__(
          self,
          vocab,
          *,
          punct_chars,
          use_endlines = None,
          ignore_excluded = True,
    ):
        if punct_chars is None:
            punct_chars = punctuation

        if use_endlines is not None:
            print("The use_endlines parameter of eds.sentences is deprecated and has been replaced by the ignore_excluded parameter")

        self.ignore_excluded = builtins.bool(ignore_excluded or use_endlines)
        self.newline_hash = vocab.strings["\n"]
        self.excluded_hash = vocab.strings["EXCLUDED"]
        self.endline_hash = vocab.strings["ENDLINE"]
        self.punct_chars_hash = {vocab.strings[c] for c in punct_chars}
        self.capitalized_shapes_hash = {
            vocab.strings[shape]
            for shape in ("Xx", "Xxx", "Xxxx", "Xxxxx")
        }

    def __call__(self, doc: spacy.tokens.Doc):
        self.process(doc)
        return doc

    cdef void process(self, Doc doc) nogil:
        """
        Segments the document in sentences.

        Arguments
        ---------
        docs: Iterable[Doc]
            A list of spacy Doc objects.
        """
        cdef TokenC token
        cdef cbool seen_period
        cdef cbool seen_newline
        cdef cbool is_in_punct_chars
        cdef cbool is_newline

        seen_period = False
        seen_newline = False

        if doc.length == 0:
            return

        for i in range(doc.length):
            # To set the attributes at False by default for the other tokens
            doc.c[i].sent_start = 1 if i == 0 else -1

            token = doc.c[i]

            if self.ignore_excluded and token.tag == self.excluded_hash:
                continue

            is_in_punct_chars = (
                  self.punct_chars_hash.const_find(token.lex.orth)
                  != self.punct_chars_hash.const_end()
            )
            is_newline = (
                  Lexeme.c_check_flag(token.lex, IS_SPACE)
                  and token.lex.orth == self.newline_hash
            )

            if seen_period or seen_newline:
                if seen_period and Lexeme.c_check_flag(token.lex, IS_DIGIT):
                    continue
                if (
                      is_in_punct_chars
                      or is_newline
                      or Lexeme.c_check_flag(token.lex, IS_PUNCT)
                ):
                    continue
                if seen_period:
                    doc.c[i].sent_start = 1
                    seen_newline = False
                    seen_period = False
                else:
                    doc.c[i].sent_start = (
                        1 if (
                              self.capitalized_shapes_hash.const_find(token.lex.shape)
                              != self.capitalized_shapes_hash.const_end()
                        ) else -1
                    )
                    seen_newline = False
                    seen_period = False
            elif is_in_punct_chars:
                seen_period = True
            elif is_newline:
                seen_newline = True



from typing import Dict, List

from loguru import logger
from spacy.tokens import Doc, Span
from spacy.util import filter_spans

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.core.matcher.matcher import GenericMatcher

from . import patterns


class SectionsMatcher(GenericMatcher):
    '''
    The `eds.sections` component extracts section titles from clinical documents.
    A "section" is then defined as the span of text between two titles.

    Here is the list of sections that are currently targeted :

    - `allergies`
    - `antécédents`
    - `antécédents familiaux`
    - `traitements entrée`
    - `conclusion`
    - `conclusion entrée`
    - `habitus`
    - `correspondants`
    - `diagnostic`
    - `données biométriques entrée`
    - `examens`
    - `examens complémentaires`
    - `facteurs de risques`
    - `histoire de la maladie`
    - `actes`
    - `motif`
    - `prescriptions`
    - `traitements sortie`
    - `evolution`
    - `modalites sortie`
    - `vaccinations`
    - `introduction`

    <!-- ![Section extraction](/resources/sections.svg){ align=right width="35%"} -->

    Remarks :

    - section `introduction` corresponds to the span of text between the header
      "COMPTE RENDU D'HOSPITALISATION" (usually denoting the beginning of the document)
      and the title of the following detected section
    - this matcher works well for hospitalization summaries (CRH), but not necessarily
      for all types of documents (in particular for emergency or scan summaries
      CR-IMAGERIE)

    !!! warning "Experimental"

        Should you rely on `eds.sections` for critical downstream tasks, make sure to
        validate the results to make sure that the component works in your case.

    Examples
    --------
    The following snippet detects section titles. It is complete and can be run _as is_.

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.sections())

    text = """
    CRU du 10/09/2021
    Motif :
    Patient admis pour suspicion de COVID
    """

    doc = nlp(text)

    doc.spans["section_titles"]
    # Out: [Motif]
    ```

    Extensions
    ----------
    The `eds.sections` matcher adds two fields to the `doc.spans` attribute :

    1. The `section_titles` key contains the list of all section titles extracted using
       the list declared in the `terms.py` module.
    2. The `sections` key contains a list of sections, ie spans of text between two
       section titles (or the last title and the end of the document).

    If the document has entities before calling this matcher an attribute `section`
    is added to each entity.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    sections : Dict[str, List[str]]
        Dictionary of terms to look for.
    attr : str
        Default attribute to match on.
    add_patterns : bool
        Whether add update patterns to match start / end of lines
    ignore_excluded : bool
        Whether to skip excluded tokens.

    Authors and citation
    --------------------
    The `eds.sections` matcher was developed by AP-HP's Data Science team.
    '''

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str = "sections",
        *,
        sections: Dict[str, List[str]] = patterns.sections,
        add_patterns: bool = True,
        attr: str = "NORM",
        ignore_excluded: bool = True,
    ):
        logger.warning(
            "The component Sections is still in Beta. Use at your own risks."
        )

        if sections is None:
            sections = patterns.sections
        sections = dict(sections)

        self.add_patterns = add_patterns
        if add_patterns:
            for k, v in sections.items():
                sections[k] = [
                    r"(?<=(?:\n|^)[^\n]{0,5})" + ent + r"(?=[^\n]{0,5}\n)" for ent in v
                ]

        super().__init__(
            nlp,
            name=name,
            terms=None,
            regex=sections,
            attr=attr,
            ignore_excluded=ignore_excluded,
            span_setter={},
        )

        self.set_extensions()

        if not nlp.has_pipe("normalizer") and not nlp.has_pipe("eds.normalizer"):
            logger.warning("You should add pipe `eds.normalizer`")

    @classmethod
    def set_extensions(cls):
        if not Span.has_extension("section_title"):
            Span.set_extension("section_title", default=None)

        if not Span.has_extension("section"):
            Span.set_extension("section", default=None)

    @classmethod
    def tag_ents(cls, sections: List[Span]):
        for section in sections:
            for e in section.ents:
                e._.section = section.label_

    # noinspection PyProtectedMember
    def __call__(self, doc: Doc) -> Doc:
        """
        Divides the doc into sections

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        doc:
            spaCy Doc object, annotated for sections
        """
        titles = filter_spans(self.process(doc))

        sections = []

        for t1, t2 in zip(titles[:-1], titles[1:]):
            section = Span(doc, t1.start, t2.start, label=t1.label)
            section._.section_title = t1
            sections.append(section)

        if titles:
            t = titles[-1]
            section = Span(doc, t.start, len(doc), label=t.label)
            section._.section_title = t
            sections.append(section)

        doc.spans["sections"] = sections
        doc.spans["section_titles"] = titles

        self.tag_ents(sections)
        return doc



from typing import Any, Dict

from typing_extensions import Literal

from edsnlp.core import PipelineProtocol, registry
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.core.terminology.terminology import TerminologyMatcher

from .patterns import get_patterns

DEFAULT_CONFIG = dict(
    attr="NORM",
    ignore_excluded=False,
    ignore_space_tokens=False,
    term_matcher="exact",
    term_matcher_config={},
    label="drug",
    span_setter={"ents": True, "drug": True},
)


@registry.factory.register(
    "eds.drugs",
    assigns=["doc.ents", "doc.spans"],
)
def create_component(
    nlp: PipelineProtocol,
    name: str = "drugs",
    *,
    attr: str = "NORM",
    ignore_excluded: bool = False,
    ignore_space_tokens: bool = False,
    term_matcher: Literal["exact", "simstring"] = "exact",
    term_matcher_config: Dict[str, Any] = {},
    label: str = "drug",
    span_setter: SpanSetterArg = {"ents": True, "drug": True},
):
    """
    The `eds.drugs` pipeline component detects mentions of French drugs (brand names and
    active ingredients) and adds them to `doc.ents`. Each drug is mapped to an
    [ATC](https://enwp.org/?curid=2770) code through the Romedi terminology
    ([@cossin:hal-02987843]). The ATC classifies drugs into groups.

    Examples
    --------
    In this example, we are looking for an oral antidiabetic medication
    (ATC code: A10B).

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.drugs(term_matcher="exact"))

    text = "Traitement habituel: Kardégic, cardensiel (bisoprolol), glucophage, lasilix"

    doc = nlp(text)

    drugs_detected = [(x.text, x.kb_id_) for x in doc.ents]

    drugs_detected[0]
    # Out: ('Kardégic', 'B01AC06')

    len(drugs_detected)
    # Out: 5

    oral_antidiabetics_detected = list(
        filter(lambda x: (x[1].startswith("A10B")), drugs_detected)
    )
    oral_antidiabetics_detected
    # Out: [('glucophage', 'A10BA02')]
    ```

    Glucophage is the brand name of a medication that contains metformine, the
    first-line medication for the treatment of type 2 diabetes.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        The name of the component
    attr : str
        The default attribute to use for matching.
    ignore_excluded : bool
        Whether to skip excluded tokens (requires an upstream
        pipeline to mark excluded tokens).
    ignore_space_tokens : bool
        Whether to skip space tokens during matching.
    term_matcher: Literal["exact", "simstring"]
        The matcher to use for matching phrases ?
        One of (exact, simstring)
    term_matcher_config: Dict[str,Any]
        Parameters of the matcher term matcher
    label : str
        Label name to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Returns
    -------
    TerminologyMatcher

    # Authors and citation

    The `eds.drugs` pipeline was developed by the IAM team and CHU de Bordeaux's Data
    Science team.
    """
    return TerminologyMatcher(
        nlp=nlp,
        name=name,
        regex=dict(),
        terms=get_patterns(),
        attr=attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
        term_matcher=term_matcher,
        term_matcher_config=term_matcher_config,
        label=label,
        span_setter=span_setter,
    )



import pytest

from edsnlp.matchers.phrase import EDSPhraseMatcher


def test_eds_phrase_matcher(doc, nlp):
    matcher = EDSPhraseMatcher(nlp.vocab, attr="TEXT")

    matcher.add("test", list(nlp.pipe(["test"])))
    matcher.remove("test")

    matcher.add("patient", list(nlp.pipe(["patient"])))

    matches = matcher(doc, as_spans=False)

    assert list(matches)

    matches = matcher(doc[:10])

    assert list(matches)


def test_offset(blank_nlp):

    text = "Ceci est un test de matching"

    doc = blank_nlp(text)
    pattern = blank_nlp("matching")

    matcher = EDSPhraseMatcher(blank_nlp.vocab, attr="TEXT")

    matcher.add("test", [pattern])

    for _, start, end in matcher(doc):
        assert doc[start:end].text == pattern.text

    for span in matcher(doc, as_spans=True):
        span.text == pattern.text

    for _, start, end in matcher(doc[2:]):
        assert doc[start:end].text == pattern.text

    for span in matcher(doc[2:], as_spans=True):
        span.text == pattern.text


def test_remove(blank_nlp):

    pattern = blank_nlp("matching")
    pattern2 = blank_nlp("Ceci")

    matcher = EDSPhraseMatcher(blank_nlp.vocab, attr="TEXT")

    matcher.add("test", [pattern])
    matcher.add("test", [pattern2])

    assert len(matcher) == 1

    with pytest.raises(KeyError):
        matcher.remove("wrong_key")

    matcher.remove("test")

    assert len(matcher) == 0



from typing import List

from pytest import fixture, mark

from edsnlp.pipelines.qualifiers.family import FamilyContext
from edsnlp.utils.examples import parse_example

examples: List[str] = [
    (
        "Le père est <ent family=True>asthmatique</ent>, "
        "sans traitement traitement particulier."
    ),
    "Son père est atteint de la <ent family=True>COVID</ent>",
    "Son père a une infection au <ent family=True>COVID</ent>",
    "Son père a une possible infection au <ent family=True>COVID</ent>",
    (
        "Le père du patient a eu un <ent family_=FAMILY>cancer du colon</ent>. "
        "La mère se porte bien."
    ),
    "Antécédents familiaux : <ent family_=FAMILY>diabète</ent>.",
    "Un <ent family_=PATIENT>relevé</ent> sanguin a été effectué.",
    (
        "Antécédent familiaux de diabète mais pas "
        "<ent family_=PATIENT>détecté</ent> jusqu'ici."
    ),
    "mère : <ent family=True>diabète de type II</ent>",
]


@fixture
def family_factory(blank_nlp):

    default_config = dict(
        family=None,
        termination=None,
        attr="NORM",
        use_sections=False,
        explain=True,
    )

    def factory(on_ents_only, **kwargs):

        config = dict(**default_config)
        config.update(kwargs)

        return FamilyContext(
            nlp=blank_nlp,
            on_ents_only=on_ents_only,
            **config,
        )

    return factory


@mark.parametrize("on_ents_only", [True, False])
@mark.parametrize("use_sections", [True, False])
def test_family(blank_nlp, family_factory, on_ents_only, use_sections):

    family = family_factory(
        on_ents_only=on_ents_only,
        use_sections=use_sections,
    )

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = family(doc)

        for entity, ent in zip(entities, doc.ents):

            for modifier in entity.modifiers:

                assert bool(ent._.family_cues) == (modifier.value in {"FAMILY", True})

                assert (
                    getattr(ent._, modifier.key) == modifier.value
                ), f"{modifier.key} labels don't match."

                if not on_ents_only:
                    for token in ent:
                        assert (
                            getattr(token._, modifier.key) == modifier.value
                        ), f"{modifier.key} labels don't match."



from edsnlp.utils.collections import (
    batch_compress_dict,
    batchify,
    decompress_dict,
    dl_to_ld,
    flatten,
    flatten_once,
    get_deep_attr,
    ld_to_dl,
    multi_tee,
    set_deep_attr,
)


def test_multi_tee():
    gen = (i for i in range(10))
    tee = multi_tee(gen)
    items1 = [value for i, value in zip(tee, range(5))]
    items2 = [value for i, value in zip(tee, range(5))]
    assert items1 == items2 == [0, 1, 2, 3, 4]

    # not the behavior I'd like (continue from 5 would be nice) but at least
    # the generator is not exhausted
    assert next(gen) == 6

    assert multi_tee(items1) is items1


def test_flatten():
    items = [1, [2, 3], [[4, 5], 6], [[[7, 8], 9], 10]]
    assert list(flatten_once(items)) == [1, 2, 3, [4, 5], 6, [[7, 8], 9], 10]
    assert list(flatten(items)) == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_dict_of_lists():
    items = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
    assert ld_to_dl(items) == {"a": [1, 3], "b": [2, 4]}
    assert list(dl_to_ld({"a": [1, 3], "b": [2, 4]})) == items


def test_dict_compression():
    list1 = [1, 2, 3]
    mapping1 = {"a": list1, "b": {"c": list1, "d": 4}}
    list2 = [1, 2, 3]
    mapping2 = {"a": list2, "b": {"c": list2, "d": 4}}

    samples = [mapping1, mapping2]
    assert list(batch_compress_dict(samples)) == [
        {"a|b/c": [1, 2, 3], "b/d": 4},
        {"a|b/c": [1, 2, 3], "b/d": 4},
    ]

    res = decompress_dict(
        [
            {"a|b/c": [1, 2, 3], "b/d": 4},
            {"a|b/c": [1, 2, 3], "b/d": 4},
        ]
    )
    assert res == {
        "a": [[1, 2, 3], [1, 2, 3]],
        "b": {"c": [[1, 2, 3], [1, 2, 3]], "d": [4, 4]},
    }
    assert res["a"] is res["b"]["c"]


def test_batchify():
    items = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    batches = list(batchify(items, 3))
    assert len(batches) == 4
    batches = list(batches)
    assert batches == [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]


def test_deep_path():
    class custom:
        def __init__(self, values):
            self.values = values

    item = {"a": {"b": (0, 1), "other": custom((1, 2))}}
    assert get_deep_attr(item, "a.b.0") == 0
    assert get_deep_attr(item, "a.other.values.0") == 1
    set_deep_attr(item, "a.b.1", 2)
    set_deep_attr(item, "a.other.values.0", 1000)
    assert item["a"]["b"] == (0, 2)
    assert item["a"]["other"].values == (1000, 2)



from typing import List

pseudo: List[str] = [
    "aucun changement",
    "aucun doute",
    "aucune hésitation",
    "aucune diminution",
    "permet pas d'",
    "permet pas de",
    "permet pas d'écarter",
    "permet pas d'éliminer",
    "permet pas d'exclure",
    "permet pas d'infirmer",
    "permet pas d'interdire",
    "permet pas de nier",
    "permet pas de réfuter",
    "permet pas de rejeter",
    "ne permet pas d'",
    "ne permet pas de",
    "ne permet pas d'écarter",
    "ne permet pas d'éliminer",
    "ne permet pas d'exclure",
    "ne permet pas d'infirmer",
    "ne permet pas d'interdire",
    "ne permet pas de nier",
    "ne permet pas de réfuter",
    "ne permet pas de rejeter",
    "n'exclut pas",
    "non négligeable",
    "pas d'amélioration",
    "pas d'augmentation",
    "pas d'autre",
    "pas de changement",
    "pas de diminution",
    "pas de doute",
    "pas écarté",
    "pas écartée",
    "pas écartées",
    "pas exclu",
    "pas exclue",
    "pas exclues",
    "pas exclus",
    "pas immunisé",
    "pas immunisée",
    "pas immunisés",
    "pas immunisées",
    "sans amélioration",
    "sans aucun doute",
    "sans augmentation",
    "sans certitude",
    "sans changement",
    "sans diminution",
    "sans doute",
    "sans être certain",
]

preceding: List[str] = [
    "à la place de",
    "absence",
    "aucun",
    "aucune",
    "aucunes",
    "aucuns",
    "décline",
    "décliné",
    "dépourvu",
    "dépourvue",
    "dépourvues",
    "dépourvus",
    "disparition de",
    "disparition des",
    "élimination",
    "exclusion",
    "excluent",
    "exclut",
    "impossibilité de",
    "immunisé",
    "immunisée",
    "immunisés",
    "immunisées",
    "incompatible avec",
    "incompatibles avec",
    "infirmation de",
    "infirmation du",
    "jamais",
    "négatif pour",
    "ni",
    "niant",
    "nie",
    "nié",
    "non",
    "nullement",
    "pas",
    "permet d'exclure",
    "peu d'argument en",
    "peu d'argument pour",
    "peu d'arguments en",
    "peu d'arguments pour",
    "plus d'aspect de",
    "sans",
    "symptôme atypique",
]

preceding_regex = [
    # ne (up to 3 words separated by spaces or newlines) pas/point/...
    r"\bne\b(?=[ \n]*(?:\w*[ \n]*){3}(?:pas|point|ni|aucun|jamais|rien))"
]

following: List[str] = [
    ":0",
    ": 0",
    ":non",
    ": non",
    "absent",
    "absente",
    "absentes",
    "absents",
    "dépourvu",
    "dépourvue",
    "dépourvues",
    "dépourvus",
    "disparaissent",
    "disparait",
    "élimination",
    "est exclu",
    "est exclue",
    "immunisé",
    "immunisée",
    "immunisés",
    "immunisées",
    "impossible",
    "improbable",
    "négatif",
    "négatifs",
    "négative",
    "négatives",
    "négligeable",
    "négligeables",
    "nié",
    "niée",
    "non",
    "pas nécessaire",
    "peu probable",
    "sont exclues",
    "sont exclus",
]

verbs: List[str] = [
    "écarter",
    "éliminer",
    "exclure",
    "infirmer",
    "interdire",
    "nier",
    "réfuter",
    "rejeter",
]



from functools import lru_cache
from typing import List, Tuple, Union

import spacy.attrs
from spacy.tokens import Doc, Span


@lru_cache(32)
def aggregate_tokens(
    doc: Doc,
    attr: str,
    ignore_excluded: bool = False,
    ignore_space_tokens: bool = False,
) -> Tuple[str, List[int], List[int], bytes]:
    """
    Aggregate tokens strings, computed from their `attr` attribute, into a single
    string, possibly ignoring excluded tokens (like pollution tokens) and/or space
    tokens. This also returns the start and end offsets of each token in the
    aggregated string, as well as a bytes array indicating which tokens were kept.
    The reason for the bytes array is that it is faster to index, and allows reverse
    indexing as well.

    Parameters
    ----------
    doc: Doc
    attr: str
    ignore_excluded: bool
    ignore_space_tokens: bool

    Returns
    -------
    Tuple[str, List[int], List[int], bytes]
        The aggregated text, the start offsets, the end offsets, and the bytes array
        indicating which tokens were kept.
    """
    idx_to_strings = doc.vocab.strings
    exclude_hash = idx_to_strings["EXCLUDED"]
    space_hash = idx_to_strings["SPACE"]
    spacy_attr = "ORTH" if attr in {"TEXT", "LOWER"} else attr
    offset = 0
    if not ignore_excluded and not ignore_space_tokens and spacy_attr == "ORTH":
        arr = doc.to_array(
            [spacy.attrs.SPACY, spacy.attrs.ORTH, spacy.attrs.IDX, spacy.attrs.LENGTH]
        )
        text_parts = [""] * len(arr)
        for i, (str_hash, space) in enumerate(
            zip(arr[:, 1].tolist(), arr[:, 0].tolist())
        ):
            text_parts[i] = idx_to_strings[str_hash] + (" " if space else "")
        begins = arr[:, 2].tolist()
        ends = (arr[:, 2] + arr[:, 3]).tolist()
        keep_list = [True] * len(arr)
    else:
        if hasattr(spacy.attrs, spacy_attr):
            arr = doc.to_array(
                [
                    spacy.attrs.SPACY,
                    spacy.attrs.TAG,
                    getattr(spacy.attrs, spacy_attr),
                ]
            )
            tokens_space = arr[:, 0].tolist()
            tokens_tag = arr[:, 1]
            tokens_text = arr[:, 2].tolist()
        # else:
        #     arr = doc.to_array([spacy.attrs.SPACY, spacy.attrs.TAG])
        #     tokens_space = arr[:, 0].tolist()
        #     tokens_tag = arr[:, 1]
        #     tokens_text = [token._.get(spacy_attr) for token in doc]

        text_parts = [""] * len(arr)
        begins = [0] * len(arr)
        ends = [0] * len(arr)
        if ignore_excluded and ignore_space_tokens:
            keep_list = (
                (tokens_tag != exclude_hash) & (tokens_tag != space_hash)
            ).tolist()
        elif ignore_excluded:
            keep_list = (tokens_tag != exclude_hash).tolist()
        elif ignore_space_tokens:
            keep_list = (tokens_tag != space_hash).tolist()
        else:
            keep_list = [True] * len(arr)

        for i, (str_hash, space, keep) in enumerate(
            zip(tokens_text, tokens_space, keep_list)
        ):
            if keep:
                if space:
                    part = idx_to_strings[str_hash] + " "
                    text_parts[i] = part
                    begins[i] = offset
                    offset += len(part)
                    ends[i] = offset - 1
                else:
                    part = idx_to_strings[str_hash]
                    text_parts[i] = part
                    begins[i] = offset
                    offset += len(part)
                    ends[i] = offset
            else:
                begins[i] = offset
                ends[i] = offset

    text = "".join(text_parts)
    if attr == "LOWER":
        text = text.lower()
    return text, begins, ends, bytes(keep_list)


def get_text(
    doclike: Union[Doc, Span],
    attr: str,
    ignore_excluded: bool,
    ignore_space_tokens: bool = False,
) -> str:
    """
    Get text using a custom attribute, possibly ignoring excluded tokens.

    Parameters
    ----------
    doclike : Union[Doc, Span]
        Doc or Span to get text from.
    attr : str
        Attribute to use.
    ignore_excluded : bool
        Whether to skip excluded tokens, by default False
    ignore_space_tokens : bool
        Whether to skip space tokens, by default False

    Returns
    -------
    str
        Extracted text.
    """
    is_doc = isinstance(doclike, Doc)
    text, starts, ends, keep = aggregate_tokens(
        doclike if is_doc else doclike.doc,
        attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
    )
    try:
        return (
            text[
                starts[keep.index(1, doclike[0].i)] : ends[
                    keep.rindex(1, None, doclike[-1].i + 1)
                ]
            ]
            if len(doclike)
            else ""
        )
    except ValueError:
        return ""


def get_char_offsets(
    doclike: Union[Doc, Span],
    attr: str,
    ignore_excluded: bool,
    ignore_space_tokens: bool = False,
) -> tuple:
    """
    Get char offsets of the doc tokens in the "cleaned" text.

    Parameters
    ----------
    doclike : Union[Doc, Span]
        Doc or Span to get text from.
    attr : str
        Attribute to use.
    ignore_excluded : bool
        Whether to skip excluded tokens, by default False
    ignore_space_tokens : bool
        Whether to skip space tokens, by default False

    Returns
    -------
    Tuple[List[int], List[int]]
        An alignment tuple: clean start/end offsets lists.
    """
    return aggregate_tokens(
        doclike if isinstance(doclike, Doc) else doclike.doc,
        attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
    )[1:3]



from edsnlp.core import registry

from .patterns import tnm_pattern
from .tnm import TNMMatcher

DEFAULT_CONFIG = dict(
    pattern=tnm_pattern,
    attr="TEXT",
    label="tnm",
    span_setter={"ents": True, "tnm": True},
)

create_component = registry.factory.register(
    "eds.tnm",
    assigns=["doc.ents", "doc.spans"],
    deprecated=["eds.TNM"],
)(TNMMatcher)



import pytest
from pytest import fixture
from spacy.tokens import Span

import edsnlp
from edsnlp.utils.span_getters import get_spans

if not Span.has_extension("label"):
    Span.set_extension("label", default=None)

if not Span.has_extension("event_type"):
    Span.set_extension("event_type", default=None)

if not Span.has_extension("test_negated"):
    Span.set_extension("test_negated", default=False)


@fixture
def gold():
    blank_nlp = edsnlp.blank("eds")
    doc1 = blank_nlp.make_doc("Arret du ttt si folfox inefficace")

    doc1.spans["sc"] = [  # drug = "folfox"
        Span(doc1, 4, 5, "drug"),  # event = "Arret"
        Span(doc1, 0, 1, "event"),  # criteria = "si"
        Span(doc1, 3, 4, "criteria"),
    ]
    doc1.spans["sc"][0]._.test_negated = False
    doc1.spans["sc"][1]._.test_negated = True
    doc1.spans["sc"][2]._.test_negated = False
    doc1.spans["sc"][1]._.event_type = "stop"

    doc1.spans["sent"] = [Span(doc1, 0, 6, "sent")]

    doc2 = blank_nlp.make_doc("Début du traitement")

    span = Span(doc2, 0, 1, "event")
    doc2.ents = [  # drug = "Début"
        span,
    ]
    span._.test_negated = False
    span._.event_type = "start"

    doc2.spans["sent"] = [Span(doc2, 0, 3, "sent")]

    return [doc1, doc2]


@pytest.mark.parametrize("with_constraints_and_not_none", [True, False])
def test_span_qualifier(gold, with_constraints_and_not_none, tmp_path):
    import torch

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        "eds.transformer",
        name="transformer",
        config=dict(
            model="prajjwal1/bert-tiny",
            window=128,
            stride=96,
        ),
    )
    nlp.add_pipe(
        "eds.span_qualifier",
        name="qualifier",
        config={
            "embedding": {
                "@factory": "eds.span_pooler",
                "embedding": nlp.get_pipe("transformer"),
                "span_getter": ["ents", "sc"],
            },
            "qualifiers": {"_.test_negated": True, "_.event_type": ("event",)}
            if with_constraints_and_not_none
            else ["_.test_negated", "_.event_type"],
            "keep_none": not with_constraints_and_not_none,
        },
    )
    qlf = nlp.get_pipe("qualifier")
    qlf.post_init(gold, set())
    if with_constraints_and_not_none:
        assert qlf.qualifiers == {"_.event_type": ["event"], "_.test_negated": True}
    else:
        assert qlf.qualifiers == {"_.event_type": True, "_.test_negated": True}
    if with_constraints_and_not_none:
        qlf.classifier.bias.data[:] = torch.tensor([0, 1000, 1000, 0])
        assert qlf.bindings == [
            ("_.test_negated", True, [False, True]),
            ("_.event_type", ["event"], ["start", "stop"]),
        ]
    else:
        qlf.classifier.bias.data[:] = torch.tensor([0, 1000, 0, 1000, 0])
        assert qlf.bindings == [
            ("_.test_negated", True, [False, True]),
            ("_.event_type", True, [None, "start", "stop"]),
        ]

    pred = qlf.pipe([doc.copy() for doc in gold])
    for doc in pred:
        for ent in get_spans(doc, qlf.embedding.span_getter):
            assert ent._.test_negated is True
            if ent.label_ == "event":
                if with_constraints_and_not_none is not None:
                    assert ent._.event_type == "start"
                else:
                    assert ent._.event_type is None

    with pytest.warns(UserWarning) as record:
        qlf.update_bindings(
            [
                ("_.test_negated", True, [False, True]),
                ("_.event_type", True, ["start", "stop"]),
                ("_.new_qualifier", True, ["test", "toast"]),
            ]
        )
    assert len(record) == 1
    assert record[0].message.args[0] == (
        "Added 2 new bindings. Consider retraining the model to learn these new "
        "bindings."
    )

    qlf.qualifiers = {"_.test_negated": True}
    assert qlf.bindings == [
        ("_.test_negated", True, [False, True]),
    ]

    nlp.to_disk(tmp_path / "model")
    nlp = edsnlp.load(tmp_path / "model")
    assert nlp.pipes.qualifier.classifier.bias.data.tolist() == [0, 1000]



HEART = [
    r"cardi",
    r"coronari",
    r"coronair",
    r"\bcd\b",
    r"\biva\d?\b",
    r"\bivp\d?\b",
    r"\bivg\d?\b",
    r"\bivd\d?\b",
    r"intra.?va\b",
    r"intra.?vp\b",
    r"intra.?vg\b",
    r"intra.?vd\b",
    r"circonflexe",
    r"\bcx\b",
    r"marginale",
    r"\bmg\b",
    r"\bdiago",
    r"\brvp\b",
    r"myocard",
    "apical",
    "septal",
    "ventricul",
    "coeur",
    "cœur",
    "auriculaire",
    "parietal",
    "septum",
]

BRAIN = [
    r"cerveau",
    r"cereb",
    r"cran",
    r"v4",
    r"m1",
    r"aica",
    r"\bpica",
    r"basilaire",
    r"polygone de willis",
    r"cercle de willis",
    r"sylvien",
    r"arachnoi",
    r"meninge",
    r"dura(?:l|ux)",
    r"puncti",
    r"front",
    r"tempo",
    r"occipi",
    r"parieta",
    # r"segment", Too generic
]

PERIPHERAL = [
    "pied",
    "main",
    r"\bmi\b",
    r"\bmig\b",
    r"\bmid\b",
    "membre",
    "jambe",
    "bras",
    "doigt",
    "digital",
    "orteil",
]

ASYMPTOMATIC = [
    r"asympto",
    r"sans.decompens",
    r"non.decompens",
]



import pytest

from edsnlp.pipelines.misc.consultation_dates import factory  # noqa

TEXT = """
Références : AMO/AMO
Objet : Compte-Rendu de Consultation du 07/10/2018
Madame BEESLY Pamela, âgée de 45 ans, née le 05/10/1987, a été vue en consultation
dans le service de NCK CS RHUMATO. Tel: 01-02-03-04-05

####

CR CS 3-1-2019 1/2

####

Paris, le 24 janvier 2020

####

Document signé le 10/02/2020

"""

cons = dict(
    additional_params=dict(),
    result=[
        dict(year=2018, month=10, day=7),
        dict(year=2019, month=1, day=3),
    ],
)

cons_town = dict(
    additional_params=dict(town_mention=True),
    result=[
        dict(year=2018, month=10, day=7),
        dict(year=2019, month=1, day=3),
        dict(year=2020, month=1, day=24),
    ],
)

cons_town_doc = dict(
    additional_params=dict(town_mention=True, document_date_mention=True),
    result=[
        dict(year=2018, month=10, day=7),
        dict(year=2019, month=1, day=3),
        dict(year=2020, month=1, day=24),
        dict(year=2020, month=2, day=10),
    ],
)


@pytest.mark.parametrize("date_pipeline", [True, False])
@pytest.mark.parametrize("example", [cons, cons_town, cons_town_doc])
def test_cons_dates(date_pipeline, example, blank_nlp):

    blank_nlp.add_pipe(
        "eds.normalizer",
        config=dict(lowercase=True, accents=True, quotes=True, pollution=False),
    )

    blank_nlp.add_pipe(
        "eds.consultation_dates", config=dict(**example["additional_params"])
    )

    if date_pipeline:
        blank_nlp.add_pipe("eds.dates")

    doc = blank_nlp(TEXT)

    assert not date_pipeline or len(doc.spans["dates"]) == 5

    assert len(doc.spans["consultation_dates"]) == len(example["result"])

    for span, result in zip(doc.spans["consultation_dates"], example["result"]):
        assert all([span._.consultation_date.dict()[k] == result[k] for k in result])



import warnings
from pathlib import Path
from typing import List, Optional, Union

from spacy.tokens import Doc

from edsnlp.core import PipelineProtocol
from edsnlp.data.converters import AttributesMappingArg
from edsnlp.data.standoff import (
    dump_standoff_file,
    parse_standoff_file,
    read_standoff,
    write_standoff,
)
from edsnlp.utils.span_getters import (
    SpanSetterArg,
    validate_span_setter,
)


class BratConnector(object):
    """
    Deprecated. Use `edsnlp.data.read_standoff` and `edsnlp.data.write_standoff`
    instead.
    Two-way connector with BRAT. Supports entities only.

    Parameters
    ----------
    directory : Union[str, Path]
        Directory containing the BRAT files.
    n_jobs : int, optional
        Number of jobs for multiprocessing, by default 1
    attributes: Optional[Union[Sequence[str], Mapping[str, str]]]
        Mapping from BRAT attributes to spaCy Span extensions.
        Extensions / attributes that are not in the mapping are not imported or exported
        If left to None, the mapping is filled with all BRAT attributes.
    span_groups: Optional[Sequence[str]]
        Additional span groups to look for entities in spaCy documents when exporting.
        Missing label (resp. span group) names are not imported (resp. exported)
        If left to None, the sequence is filled with all BRAT entity labels.
    """

    def __init__(
        self,
        directory: Union[str, Path],
        n_jobs: int = 1,
        attributes: Optional[AttributesMappingArg] = None,
        bool_attributes: Optional[List[str]] = [],
        span_groups: SpanSetterArg = ["ents", "*"],
        keep_raw_attribute_values: bool = False,
    ):
        warnings.warn(
            "This connector is deprecated and will be removed in a future version.\n"
            "Use `edsnlp.data.read_standoff` and `edsnlp.data.write_standoff` instead.",
            DeprecationWarning,
        )
        self.directory: Path = Path(directory)
        self.attr_map = attributes
        self.span_setter = validate_span_setter(span_groups)
        self.keep_raw_attribute_values = keep_raw_attribute_values
        self.bool_attributes = list(bool_attributes)

    def brat2docs(self, nlp: PipelineProtocol, run_pipe=False) -> List[Doc]:
        res = read_standoff(
            path=self.directory,
            nlp=nlp,
            keep_txt_only_docs=True,
            span_attributes=self.attr_map,
            span_setter=self.span_setter,
            keep_raw_attribute_values=self.keep_raw_attribute_values,
            bool_attributes=self.bool_attributes,
        )
        return list(nlp.pipe(res) if run_pipe else res)

    def docs2brat(self, docs: List[Doc]) -> None:
        """
        Writes a list of spaCy documents to file.
        """
        write_standoff(
            docs,
            self.directory,
            span_getter=self.span_setter,
            span_attributes=self.attr_map or {},
            overwrite=True,
        )


load_from_brat = parse_standoff_file
export_to_brat = dump_standoff_file



results_tobacco = dict(
    has_match=[
        True,
        True,
        False,
        True,
        True,
        True,
        True,
    ],
    detailled_status=[
        "PRESENT",
        "PRESENT",
        None,
        "ABSTINENCE",
        "ABSENT",
        "ABSENT",
        "ABSTINENCE",
    ],
    assign=[{"PA": 15}] + 6 * [None],
    texts=[
        "Tabagisme évalué à 15 PA",
        "Patient tabagique",
        "Tabagisme festif",
        "On a un tabagisme ancien",
        "Tabac: 0",
        "Tabagisme passif",
        "Tabac: sevré depuis 5 ans",
    ],
)



"""`eds.solid_tumor` pipeline"""
from typing import Any, Dict, List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.utils.numbers import parse_digit

from ..base import DisorderMatcher
from .patterns import default_patterns


class SolidTumorMatcher(DisorderMatcher):
    """
    The `eds.solid_tumor` pipeline component extracts mentions of solid tumors. It will
    notably match:

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/disorders/solid_tumor/patterns.py"
        # fmt: on
        ```

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: set to either
        - `"METASTASIS"` for tumors at the metastatic stage
        - `"LOCALIZED"` else
    - `span._.assigned`: dictionary with the following keys, if relevant:
        - `stage`: stage of the tumor

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(eds.solid_tumor())
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/solid-tumor-examples.md"

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline
    name : Optional[str]
        The name of the component
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc
    use_tnm : bool
        Whether to use TNM scores matching as well

    Authors and citation
    --------------------
    The `eds.solid_tumor` component was developed by AP-HP's Data Science team with a
    team of medical experts. A paper describing in details the development of those
    components is being drafted and will soon be available.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "solid_tumor",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]] = default_patterns,
        use_tnm: bool = False,
        label: str = "solid_tumor",
        span_setter: SpanSetterArg = {"ents": True, "solid_tumor": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            patterns=patterns,
            detailed_status_mapping={
                0: "ABSENT",
                1: "LOCALIZED",
                2: "METASTASIS",
            },
            label=label,
            span_setter=span_setter,
        )

        self.use_tnm = use_tnm

        if use_tnm:
            from edsnlp.pipes.ner.tnm import TNM

            self.tnm = TNM(nlp, pattern=None, attr="TEXT")

    def process_tnm(self, doc):
        spans = self.tnm.process(doc)
        spans = self.tnm.parse(spans)

        for span in spans:
            span.label_ = "solid_tumor"
            span._.source = "tnm"
            metastasis = span._.value.dict().get("metastasis", "0")
            if metastasis == "1":
                span._.status = 2
            yield span

    def process(self, doc: Doc) -> List[Span]:
        for span in super().process(doc):
            if (span._.source == "metastasis") or (
                "metastasis" in span._.assigned.keys()
            ):
                span._.status = 2

            if "stage" in span._.assigned.keys():
                stage = parse_digit(
                    span._.assigned["stage"],
                    attr="NORM",
                    ignore_excluded=True,
                )
                if stage == 4:
                    span._.status = 2

            yield span

        if self.use_tnm:
            yield from self.process_tnm(doc)



"""`eds.dates` pipeline."""
import warnings
from itertools import chain
from typing import Dict, Iterable, List, Optional, Tuple, Union

from confit import VisibleDeprecationWarning
from loguru import logger
from spacy.tokens import Doc, Span
from typing_extensions import Literal

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.pipes.base import (
    BaseNERComponent,
    SpanGetterArg,
    SpanSetterArg,
    get_spans,
    validate_span_getter,
)
from edsnlp.utils.filter import align_spans, filter_spans

from . import patterns
from .models import AbsoluteDate, Bound, Duration, Mode, Period, RelativeDate


class DatesMatcher(BaseNERComponent):
    """
    The `eds.dates` matcher detects and normalize dates within a medical document.
    We use simple regular expressions to extract date mentions.

    Scope
    -----
    The `eds.dates` pipeline finds absolute (eg `23/08/2021`) and relative (eg `hier`,
    `la semaine dernière`) dates alike. It also handles mentions of duration.

    | Type       | Example                       |
    | ---------- | ----------------------------- |
    | `absolute` | `3 mai`, `03/05/2020`         |
    | `relative` | `hier`, `la semaine dernière` |
    | `duration` | `pendant quatre jours`        |

    See the [tutorial](../../tutorials/detecting-dates.md) for a presentation of a
    full pipeline featuring the `eds.dates` component.

    ## Usage

    ```python
    import edsnlp, edsnlp.pipes as eds

    import pendulum

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.dates())

    text = (
        "Le patient est admis le 23 août 2021 pour une douleur à l'estomac. "
        "Il lui était arrivé la même chose il y a un an pendant une semaine. "
        "Il a été diagnostiqué en mai 1995."
    )

    doc = nlp(text)

    dates = doc.spans["dates"]
    dates
    # Out: [23 août 2021, il y a un an, mai 1995]

    dates[0]._.date.to_datetime()
    # Out: 2021-08-23T00:00:00+02:00

    dates[1]._.date.to_datetime()
    # Out: None

    note_datetime = pendulum.datetime(2021, 8, 27, tz="Europe/Paris")

    dates[1]._.date.to_datetime(note_datetime=note_datetime)
    # Out: 2020-08-27T00:00:00+02:00

    date_2_output = dates[2]._.date.to_datetime(
        note_datetime=note_datetime,
        infer_from_context=True,
        tz="Europe/Paris",
        default_day=15,
    )
    date_2_output
    # Out: 1995-05-15T00:00:00+02:00

    doc.spans["durations"]
    # Out: [pendant une semaine]
    ```

    Extensions
    ----------
    The `eds.dates` pipeline declares two extensions on the `Span` object:

    - the `span._.date` attribute of a date contains a parsed version of the date.
    - the `span._.duration` attribute of a duration contains a parsed version of the
      duration.

    As with other components, you can use the `span._.value` attribute to get either the
    parsed date or the duration depending on the span.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : Optional[str]
        Name of the pipeline component
    absolute : Union[List[str], str]
        List of regular expressions for absolute dates.
    relative : Union[List[str], str]
        List of regular expressions for relative dates
        (eg `hier`, `la semaine prochaine`).
    duration : Union[List[str], str]
        List of regular expressions for durations
        (eg `pendant trois mois`).
    false_positive : Union[List[str], str]
        List of regular expressions for false positive (eg phone numbers, etc).
    span_getter : SpanGetterArg
        Where to look for dates in the doc. By default, look in the whole doc. You can
        combine this with the `merge_mode` argument for interesting results.
    merge_mode : Literal["intersect", "align"]
        How to merge matched dates with the spans from `span_getter`, if given:

        - `intersect`: return only the matches that fall in the `span_getter` spans
        - `align`: if a date overlaps a span from `span_getter` (e.g. a date extracted
          by a machine learning model), return the `span_getter` span instead, and
          assign all the parsed information (`._.date` / `._.duration`) to it. Otherwise
          don't return the date.
    on_ents_only : Union[bool, str, Iterable[str]]
        Deprecated, use `span_getter` and `merge_mode` instead.
        Whether to look on dates in the whole document or in specific sentences:

        - If `True`: Only look in the sentences of each entity in doc.ents
        - If False: Look in the whole document
        - If given a string `key` or list of string: Only look in the sentences of
          each entity in `#!python doc.spans[key]`
    detect_periods : bool
        Whether to detect periods (experimental)
    detect_time: bool
        Whether to detect time inside dates
    period_proximity_threshold : int
        Max number of words between two dates to extract a period.
    as_ents : bool
        Deprecated, use span_setter instead.
        Whether to treat dates as entities
    attr : str
        spaCy attribute to use
    date_label : str
        Label to use for dates
    duration_label : str
        Label to use for durations
    period_label : str
        Label to use for periods
    span_setter : SpanSetterArg
        How to set matches in the doc.

    Authors and citation
    --------------------
    The `eds.dates` pipeline was developed by AP-HP's Data Science team.
    """

    # noinspection PyProtectedMember
    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str = "dates",
        *,
        absolute: Optional[List[str]] = None,
        relative: Optional[List[str]] = None,
        duration: Optional[List[str]] = None,
        false_positive: Optional[List[str]] = None,
        on_ents_only: Optional[Union[bool, str, Iterable[str]]] = None,
        span_getter: Optional[SpanGetterArg] = None,
        merge_mode: Literal["intersect", "align"] = "intersect",
        detect_periods: bool = False,
        detect_time: bool = True,
        period_proximity_threshold: int = 3,
        as_ents: bool = False,
        attr: str = "LOWER",
        date_label: str = "date",
        duration_label: str = "duration",
        period_label: str = "period",
        span_setter: SpanSetterArg = {
            "dates": ["date"],
            "durations": ["duration"],
            "periods": ["period"],
        },
    ):
        self.date_label = date_label
        self.duration_label = duration_label
        self.period_label = period_label

        # Backward compatibility
        if as_ents is True:
            warnings.warn(
                "The `as_ents` argument is deprecated."
                + (
                    " Pass `span_setter={} instead.".format(
                        {**span_setter, "ents": [self.date_label, self.duration_label]}
                    )
                    if isinstance(span_setter, dict)
                    else " Use the `span_setter` argument instead."
                ),
                VisibleDeprecationWarning,
            )
            span_setter = dict(span_setter)
            span_setter["ents"] = True
        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

        if absolute is None:
            if detect_time:
                absolute = patterns.absolute_pattern_with_time
            else:
                absolute = patterns.absolute_pattern
        if relative is None:
            relative = patterns.relative_pattern
        if duration is None:
            duration = patterns.duration_pattern
        if false_positive is None:
            false_positive = patterns.false_positive_pattern

        if isinstance(absolute, str):
            absolute = [absolute]
        if isinstance(relative, str):
            relative = [relative]
        if isinstance(duration, str):
            relative = [duration]
        if isinstance(false_positive, str):
            false_positive = [false_positive]

        if on_ents_only is None and span_getter is None:
            on_ents_only = False

        if on_ents_only:
            assert span_getter is None, (
                "Cannot use both `on_ents_only` and " "`span_getter`"
            )

            def span_getter(doc):
                return (span.sent for span in doc.ents)

            merge_mode = "intersect"
            warnings.warn(
                "The `on_ents_only` argument is deprecated."
                " Use the `span_getter` argument instead.",
                VisibleDeprecationWarning,
            )
        self.span_getter = validate_span_getter(span_getter, optional=True)
        self.merge_mode = merge_mode
        self.regex_matcher = RegexMatcher(attr=attr, alignment_mode="strict")

        self.regex_matcher.add("false_positive", false_positive)
        self.regex_matcher.add("absolute", absolute)
        self.regex_matcher.add("relative", relative)
        self.regex_matcher.add("duration", duration)

        self.detect_periods = detect_periods
        self.period_proximity_threshold = period_proximity_threshold

        self.as_ents = as_ents

        if detect_periods:
            logger.warning("The period extractor is experimental.")

        self.set_extensions()

    def set_extensions(self) -> None:
        """
        Set extensions for the dates pipeline.
        """

        if not Span.has_extension("datetime"):
            Span.set_extension("datetime", default=None)

        if not Span.has_extension(self.date_label):
            Span.set_extension(self.date_label, default=None)

        if not Span.has_extension(self.duration_label):
            Span.set_extension(self.duration_label, default=None)

        if not Span.has_extension(self.period_label):
            Span.set_extension(self.period_label, default=None)

    def process(self, doc: Doc) -> List[Tuple[Span, Dict[str, str]]]:
        """
        Find dates in doc.

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        dates:
            list of date spans
        """

        spans = None

        if self.span_getter is not None:
            spans = list(get_spans(doc, self.span_getter))
            matches = []
            for sent in set([ent.sent for ent in doc]):
                matches = chain(
                    matches,
                    self.regex_matcher(
                        sent,
                        as_spans=True,
                        return_groupdict=True,
                    ),
                )

        else:
            matches = self.regex_matcher(
                doc,
                as_spans=True,
                return_groupdict=True,
            )

        matches = filter_spans(matches)
        matches = [date for date in matches if date[0].label_ != "false_positive"]

        matches = list(self.parse(matches))

        if self.span_getter is not None:
            if self.merge_mode == "align":
                alignments = align_spans(matches, spans, sort_by_overlap=True)
                matches = []
                for span, aligned in zip(spans, alignments):
                    if len(aligned):
                        old = aligned[0]
                        span.label_ = old.label_
                        span._.set(self.date_label, old._.get(self.date_label))
                        span._.set(self.duration_label, old._.get(self.duration_label))
                        matches.append(span)

            elif self.merge_mode == "intersect":
                alignments = align_spans(matches, spans)
                matches = []
                for span, aligned in zip(spans, alignments):
                    matches.extend(aligned)
                matches = list(dict.fromkeys(matches))

        if self.detect_periods:
            matches.extend(self.process_periods(matches))

        return matches

    def parse(
        self, matches: List[Tuple[Span, Dict[str, str]]]
    ) -> Tuple[List[Span], List[Span]]:
        """
        Parse dates/durations using the groupdict returned by the matcher.

        Parameters
        ----------
        matches : List[Tuple[Span, Dict[str, str]]]
            List of tuples containing the spans and groupdict
            returned by the matcher.

        Returns
        -------
        Tuple[List[Span], List[Span]]
            List of processed spans, with the date parsed.
        """

        for span, groupdict in matches:
            date_cfg = groupdict
            for k, v in list(groupdict.items()):
                if v is not None and "_" in k:
                    key, value = k.split("_")
                    date_cfg.update({key: value})
            if span.label_ == "relative":
                parsed = RelativeDate.parse_obj(date_cfg)
                span.label_ = self.date_label
                span._.date = parsed
            elif span.label_ == "absolute":
                parsed = AbsoluteDate.parse_obj(date_cfg)
                span.label_ = self.date_label
                span._.date = parsed
            else:
                parsed = Duration.parse_obj(date_cfg)
                span.label_ = self.duration_label
                span._.duration = parsed

        return [span for span, _ in matches]

    def process_periods(self, dates: List[Span]) -> List[Span]:
        """
        Experimental period detection.

        Parameters
        ----------
        dates : List[Span]
            List of detected dates.

        Returns
        -------
        List[Span]
            List of detected periods.
        """

        if len(dates) < 2:
            return []

        periods = []
        seen = set()

        dates = list(sorted(dates, key=lambda d: d.start))

        for d1, d2 in zip(dates[:-1], dates[1:]):
            v1 = d1._.date if d1.label_ == self.date_label else d1._.duration
            v2 = d2._.date if d2.label_ == self.date_label else d2._.duration
            if v1.mode == Mode.DURATION or v2.mode == Mode.DURATION:
                pass
            elif d1 in seen or v1.bound is None or v2.bound is None:
                continue

            if (
                d1.end - d2.start < self.period_proximity_threshold
                and v1.bound != v2.bound
            ):
                period = Span(d1.doc, d1.start, d2.end, label=self.period_label)

                # If one date is a duration,
                # the other may not have a registered bound attribute.
                if v1.mode == Mode.DURATION:
                    m1 = Bound.FROM if v2.bound == Bound.UNTIL else Bound.UNTIL
                    m2 = v2.mode or Bound.FROM
                elif v2.mode == Mode.DURATION:
                    m1 = v1.mode or Bound.FROM
                    m2 = Bound.FROM if v1.bound == Bound.UNTIL else Bound.UNTIL
                else:
                    m1 = v1.mode or Bound.FROM
                    m2 = v2.mode or Bound.FROM

                period._.set(
                    self.period_label,
                    Period.parse_obj(
                        {
                            m1: d1,
                            m2: d2,
                        }
                    ),
                )

                seen.add(d1)
                seen.add(d2)

                periods.append(period)

        return periods

    def __call__(self, doc: Doc) -> Doc:
        """
        Tags dates.

        Parameters
        ----------
        doc : Doc
            spaCy Doc object

        Returns
        -------
        doc : Doc
            spaCy Doc object, annotated for dates
        """

        matches = self.process(doc)

        self.set_spans(doc, matches)

        return doc



# noinspection SpellCheckingInspection
information = [
    (
        r"(?s)(=====+\s*)?(L\s*e\s*s\sdonnées\s*administratives,\s*sociales\s*|"
        r"I?nfo\s*rmation\s*aux?\s*patients?|"
        r"L[’']AP-HP\s*collecte\s*vos\s*données\s*administratives|"
        r"L[’']Assistance\s*Publique\s*-\s*Hôpitaux\s*de\s*Paris\s*"
        r"\(?AP-HP\)?\s*a\s*créé\s*une\s*base\s*de\s*données)"
        r".{,2000}https?:\/\/recherche\.aphp\.fr\/eds\/droit-opposition[\s\.]*"
    ),
    (
        r"(?si)l’arrêt\s*du\s*tabac\s*permet\s*de\s*diminuer\s*le\s*risque\s*"
        r"de\s*maladie\s*cardiovasculaire."
    ),
]
# Example : NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbW...
bars = r"(?i)([nbw]|_|-|=){5,}"

# Biology tables: Prone to false positive with disease names
biology = r"(\b.*[|¦].*\n)+"

# Leftside note with doctor names
doctors = r"(?mi)(^((dr)|(pr))(\.|\s|of).*)+"

# Mails or websites
web = [
    r"(www\.\S*)",
    r"(\S*@\S*)",
    r"\S*\.(?:fr|com|net|org)",
]

# Subsection with ICD-10 Codes
coding = r".*? \(\d+\) [a-zA-Z]\d{2,4}.*?(\n|[a-zA-Z]\d{2,4})"


# New page
date = r"\b\d\d/\d\d/\d\d\d\d\b"
ipp = r"80\d{8}"
page = r"((^\d\/\d\s?)|(^\d\d?\/\d\d\?))"
footer = rf"(?i)({page}.*\n?pat.*(ipp)?.*\n?(courrier valid.*)?)"
footer += rf"|(.*{date}.*{ipp}.*)|(imprim.\sle\s{date}.*\d/\d.*\n?pat.*{date})"

pollution = dict(
    information=information,
    bars=bars,
    biology=biology,
    doctors=doctors,
    web=web,
    coding=coding,
    footer=footer,
)

default_enabled = dict(
    information=True,
    bars=True,
    biology=False,
    doctors=True,
    web=True,
    coding=False,
    footer=True,
)



from itertools import chain

import spacy
from pytest import fixture, raises
from spacy.tokens.span import Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipelines.misc.measurements import MeasurementsMatcher

text = (
    "Le patient fait 1 m 50 kg. La tumeur fait 2.0cm x 3cm. \n"
    "Une autre tumeur plus petite fait 2 par 1mm.\n"
    "Les trois éléments font 8, 13 et 15dm."
)


@fixture
def blank_nlp():
    model = spacy.blank("eds")
    model.add_pipe("eds.normalizer")
    model.add_pipe("eds.sentences")
    return model


@fixture
def matcher(blank_nlp: PipelineProtocol):
    return MeasurementsMatcher(
        blank_nlp,
        extract_ranges=True,
    )


def test_default_factory(blank_nlp: PipelineProtocol):
    blank_nlp.add_pipe("matcher", config=dict(terms={"patient": "patient"}))
    blank_nlp.add_pipe(
        "eds.measurements",
        config=dict(measurements=["size", "weight", "bmi"]),
    )

    doc = blank_nlp(text)

    assert len(doc.ents) == 1

    assert len(doc.spans["measurements"]) == 9


def test_measurements_component(
    blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher
):
    doc = blank_nlp(text)

    with raises(KeyError):
        doc.spans["measurements"]

    doc = matcher(doc)

    m1, m2, m3, m4, m5, m6, m7, m8, m9 = doc.spans["measurements"]

    assert str(m1._.value) == "1 m"
    assert str(m2._.value) == "50 kg"
    assert str(m3._.value) == "2.0 cm"
    assert str(m4._.value) == "3 cm"
    assert str(m5._.value) == "2 mm"
    assert str(m6._.value) == "1 mm"
    assert str(m7._.value) == "8 dm"
    assert str(m8._.value) == "13 dm"
    assert str(m9._.value) == "15 dm"


def test_measurements_component_scaling(
    blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher
):
    doc = blank_nlp(text)

    with raises(KeyError):
        doc.spans["measurements"]

    doc = matcher(doc)

    m1, m2, m3, m4, m5, m6, m7, m8, m9 = doc.spans["measurements"]

    assert m1._.value.cm == 100
    assert m2._.value.mg == 50000000.0
    assert m3._.value.mm == 20
    assert m4._.value.mm == 30
    assert m5._.value.cm == 0.2
    assert m6._.value.cm == 0.1
    assert m7._.value.dm == 8.0
    assert m8._.value.m == 1.3
    assert m9._.value.m == 1.5


def test_measure_label(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    doc = blank_nlp(text)
    doc = matcher(doc)

    m1, m2, m3, m4, m5, m6, m7, m8, m9 = doc.spans["measurements"]

    assert m1.label_ == "size"
    assert m2.label_ == "weight"
    assert m3.label_ == "size"
    assert m4.label_ == "size"
    assert m5.label_ == "size"
    assert m6.label_ == "size"
    assert m7.label_ == "size"
    assert m8.label_ == "size"
    assert m9.label_ == "size"


def test_measure_str(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    for text, res in [
        ("1m50", "1.5 m"),
        ("1,50cm", "1.5 cm"),
    ]:
        doc = blank_nlp(text)
        doc = matcher(doc)

        assert str(doc.spans["measurements"][0]._.value) == res


def test_measure_repr(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    for text, res in [
        (
            "1m50",
            "Measurement(1.5, 'm')",
        ),
        (
            "1,50cm",
            "Measurement(1.5, 'cm')",
        ),
    ]:
        doc = blank_nlp(text)
        doc = matcher(doc)

        print(doc.spans["measurements"])

        assert repr(doc.spans["measurements"][0]._.value) == res


def test_compare(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    m1, m2 = "1m0", "120cm"
    m1 = matcher(blank_nlp(m1)).spans["measurements"][0]
    m2 = matcher(blank_nlp(m2)).spans["measurements"][0]
    assert m1._.value <= m2._.value
    assert m2._.value > m1._.value

    m3 = "Entre deux et trois metres"
    m4 = "De 2 à 3 metres"
    m3 = matcher(blank_nlp(m3)).spans["measurements"][0]
    m4 = matcher(blank_nlp(m4)).spans["measurements"][0]
    print(blank_nlp("Entre deux et trois metres"))
    assert str(m3._.value) == "2-3 m"
    assert str(m4._.value) == "2-3 m"
    assert m4._.value.cm == (200.0, 300.0)

    assert m3._.value == m4._.value
    assert m3._.value <= m4._.value
    assert m3._.value >= m1._.value

    assert max(list(chain(m1._.value, m2._.value, m3._.value, m4._.value))).cm == 300


def test_unitless(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    for text, res in [
        ("BMI: 24 .", "24 kg_per_m2"),
        ("Le patient mesure 1.5 ", "1.5 m"),
        ("Le patient mesure 152 ", "152 cm"),
        ("Le patient pèse 34 ", "34 kg"),
    ]:
        doc = blank_nlp(text)
        doc = matcher(doc)

        assert str(doc.spans["measurements"][0]._.value) == res


def test_non_matches(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    for text in [
        "On délivre à 10 g / h.",
        "Le patient grandit de 10 cm par jour ",
        "Truc 10cma truc",
        "01.42.43.56.78 m",
    ]:
        doc = blank_nlp(text)
        print(list(doc))
        doc = matcher(doc)

        assert len(doc.spans["measurements"]) == 0


def test_numbers(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    for text, res in [
        ("deux m", "2 m"),
        ("2 m", "2 m"),
        ("⅛ m", "0.125 m"),
        ("0 m", "0 m"),
    ]:
        doc = blank_nlp(text)
        doc = matcher(doc)

        assert str(doc.spans["measurements"][0]._.value) == res


def test_ranges(blank_nlp: PipelineProtocol, matcher: MeasurementsMatcher):
    for text, res, snippet in [
        ("Le patient fait entre 1 et 2m", "1-2 m", "entre 1 et 2m"),
        ("On mesure de 2 à 2.5 dl d'eau", "2-2.5 dl", "de 2 à 2.5 dl"),
    ]:
        doc = blank_nlp(text)
        doc = matcher(doc)

        measurement = doc.spans["measurements"][0]
        print(doc.spans["measurements"])
        assert str(measurement._.value) == res
        assert measurement.text == snippet


def test_merge_align(blank_nlp, matcher):
    matcher.merge_mode = "align"
    matcher.span_getter = {"candidates": True}
    matcher.span_setter = {"ents": True}
    doc = blank_nlp(text)
    ent = Span(doc, 10, 15, label="size")
    doc.spans["candidates"] = [ent]
    doc = matcher(doc)

    assert len(doc.ents) == 1
    assert str(ent._.value) == "2.0 cm"


def test_merge_intersect(blank_nlp, matcher: MeasurementsMatcher):
    matcher.merge_mode = "intersect"
    matcher.span_setter = {**matcher.span_setter, "ents": True}
    matcher.span_getter = {"lookup_zones": True}
    doc = blank_nlp(text)
    ent = Span(doc, 10, 16, label="size")
    doc.spans["lookup_zones"] = [ent]
    doc = matcher(doc)

    assert len(doc.ents) == 2
    assert len(doc.spans["measurements"]) == 2
    assert [doc.ents[0].text, doc.ents[1].text] == ["2.0cm", "3cm"]
    assert [doc.ents[0]._.value.cm, doc.ents[1]._.value.cm] == [2.0, 3]


def test_measurement_snippets(blank_nlp, matcher: MeasurementsMatcher):
    for text, result in [
        ("0.50g", ["0.5 g"]),
        ("0.050g", ["0.05 g"]),
        ("1 m 50", ["1.5 m"]),
        ("1.50 m", ["1.5 m"]),
        ("1,50m", ["1.5 m"]),
        ("2.0cm x 3cm", ["2.0 cm", "3 cm"]),
        ("2 par 1mm", ["2 mm", "1 mm"]),
        ("8, 13 et 15dm", ["8 dm", "13 dm", "15 dm"]),
        ("1 / 50  kg", ["0.02 kg"]),
    ]:
        doc = blank_nlp(text)
        doc = matcher(doc)

        assert [str(span._.value) for span in doc.spans["measurements"]] == result



from edsnlp import registry

from .copd import COPDMatcher
from .patterns import default_patterns

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="copd",
    span_setter={"ents": True, "copd": True},
)

create_component = registry.factory.register(
    "eds.copd",
    assigns=["doc.ents", "doc.spans"],
    deprecated=["eds.COPD"],
)(COPDMatcher)



import pytest
from spacy.tokens import Span

from edsnlp.data.converters import (
    FILENAME,
    get_dict2doc_converter,
    get_doc2dict_converter,
)


@pytest.fixture(autouse=True, scope="module")
def set_extensions():
    if not Span.has_extension("negation"):
        Span.set_extension("negation", default=None)


def test_read_omop_dict(blank_nlp):
    json = {
        "note_id": 1234,
        "note_text": "This is a test.",
        "entities": [
            {
                "note_nlp_id": 0,
                "start_char": 0,
                "end_char": 4,
                "lexical_variant": "This",
                "note_nlp_source_value": "test",
                "negation": True,
            },
            {
                "note_nlp_id": 1,
                "start_char": 5,
                "end_char": 7,
                "lexical_variant": "is",
                "note_nlp_source_value": "test",
            },
        ],
    }
    doc = get_dict2doc_converter(
        "omop",
        dict(
            nlp=blank_nlp,
            span_attributes="negation",
            bool_attributes="negation",
        ),
    )[0](json)
    assert doc.text == "This is a test."
    assert doc._.note_id == 1234
    assert len(doc.ents) == 2
    assert doc.ents[0].text == "This"
    assert doc.ents[0]._.negation is True
    assert doc.ents[1]._.negation is False


def test_read_standoff_dict(blank_nlp):
    json = {
        "doc_id": 1234,
        "text": "This is a test.",
        "entities": [
            {
                "entity_id": 0,
                "fragments": [
                    {
                        "begin": 0,
                        "end": 4,
                    }
                ],
                "attributes": {
                    "negation": True,
                },
                "label": "test",
            },
            {
                "entity_id": 1,
                "fragments": [
                    {
                        "begin": 5,
                        "end": 7,
                    }
                ],
                "attributes": {},
                "label": "test",
            },
        ],
    }
    doc = get_dict2doc_converter(
        "standoff",
        dict(
            nlp=blank_nlp,
            span_attributes={"negation": "negation"},
            bool_attributes="negation",
        ),
    )[0](json)
    assert doc.text == "This is a test."
    assert doc._.note_id == 1234
    assert len(doc.ents) == 2
    assert doc.ents[0].text == "This"
    assert doc.ents[0]._.negation is True
    assert doc.ents[1]._.negation is False


def test_write_omop_dict(blank_nlp):
    doc = blank_nlp("This is a test.")
    doc._.note_id = 1234
    doc.ents = [Span(doc, 0, 1, label="test"), Span(doc, 1, 2, label="test")]
    doc.ents[0]._.negation = True
    doc.ents[1]._.negation = False
    json = {
        FILENAME: 1234,
        "note_id": 1234,
        "note_text": "This is a test.",
        "entities": [
            {
                "note_nlp_id": 0,
                "start_char": 0,
                "end_char": 4,
                "lexical_variant": "This",
                "note_nlp_source_value": "test",
                "sent.text": "This is a test.",
                "negation": True,
            },
            {
                "note_nlp_id": 1,
                "start_char": 5,
                "end_char": 7,
                "lexical_variant": "is",
                "note_nlp_source_value": "test",
                "sent.text": "This is a test.",
                "negation": False,
            },
        ],
    }
    assert (
        get_doc2dict_converter(
            "omop",
            dict(
                span_getter={"ents": True},
                span_attributes=["negation", "sent.text"],
            ),
        )[0](doc)
        == json
    )


def test_write_standoff_dict(blank_nlp):
    doc = blank_nlp("This is a test.")
    doc._.note_id = 1234
    doc.ents = [Span(doc, 0, 1, label="test"), Span(doc, 1, 2, label="test")]
    if not Span.has_extension("negation"):
        Span.set_extension("negation", default=None)
    doc.ents[0]._.negation = True
    doc.ents[1]._.negation = False
    json = {
        FILENAME: 1234,
        "doc_id": 1234,
        "text": "This is a test.",
        "entities": [
            {
                "entity_id": 0,
                "fragments": [
                    {
                        "begin": 0,
                        "end": 4,
                    }
                ],
                "attributes": {
                    "negation": True,
                },
                "label": "test",
            },
            {
                "entity_id": 1,
                "fragments": [
                    {
                        "begin": 5,
                        "end": 7,
                    }
                ],
                "attributes": {
                    "negation": False,
                },
                "label": "test",
            },
        ],
    }
    assert (
        get_doc2dict_converter(
            "standoff",
            dict(
                span_getter={"ents": True},
                span_attributes={"negation": "negation"},
            ),
        )[0](doc)
        == json
    )


def test_write_ents_dict(blank_nlp):
    doc = blank_nlp("This is a test.")
    doc._.note_id = 1234
    doc.ents = [Span(doc, 0, 1, label="test"), Span(doc, 1, 2, label="test")]
    doc.ents[0]._.negation = True
    doc.ents[1]._.negation = False
    jsons = [
        {
            "note_id": 1234,
            "start": 0,
            "end": 4,
            "lexical_variant": "This",
            "label": "test",
            "span_type": "ents",
            "sent.text": "This is a test.",
            "negation": True,
        },
        {
            "note_id": 1234,
            "start": 5,
            "end": 7,
            "lexical_variant": "is",
            "label": "test",
            "span_type": "ents",
            "sent.text": "This is a test.",
            "negation": False,
        },
    ]
    assert (
        get_doc2dict_converter(
            "ents",
            dict(
                span_getter={"ents": True},
                span_attributes=["negation", "sent.text"],
            ),
        )[0](doc)
        == jsons
    )


def test_unknown_converter():
    with pytest.raises(ValueError):
        get_dict2doc_converter("test", {})

    with pytest.raises(ValueError):
        get_doc2dict_converter("test", {})


def test_callable_converter():
    raw = lambda x: x  # noqa: E731
    assert get_dict2doc_converter(raw, {}) == (raw, {})
    assert get_doc2dict_converter(raw, {}) == (raw, {})



from edsnlp.connectors.labeltool import docs2labeltool

texts = [
    "Le patient est malade",
    "Le patient n'est pas malade",
    "Le patient est peut-être malade",
    "Le patient dit qu'il est malade",
]


def test_docs2labeltool(nlp):

    modifiers = ["negated", "hypothesis", "reported_speech"]

    docs = list(nlp.pipe(texts))

    df = docs2labeltool(docs, extensions=modifiers)
    assert len(df)

    df = docs2labeltool(docs)
    assert len(df)



from itertools import chain
from typing import Dict, List, Optional, Set, Union

from loguru import logger
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.phrase import EDSPhraseMatcher
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.pipes.base import BaseComponent, SpanGetterArg, validate_span_getter


def check_normalizer(nlp: PipelineProtocol) -> None:
    components = {name: component for name, component in nlp.pipeline}
    normalizer = components.get("normalizer")

    if normalizer and not normalizer.lowercase:
        logger.warning(
            "You have chosen the NORM attribute, but disabled lowercasing "
            "in your normalisation pipeline. "
            "This WILL hurt performance : you might want to use the "
            "LOWER attribute instead."
        )


def get_qualifier_extensions(nlp: PipelineProtocol):
    """
    Check for all qualifiers present in the pipe and return its corresponding extension
    """
    return {
        name: nlp.get_pipe_meta(name).assigns[0].split("span.")[-1]
        for name, pipe in nlp.pipeline
        if isinstance(pipe, RuleBasedQualifier)
    }


class RuleBasedQualifier(BaseComponent):
    """
    Implements the NegEx algorithm.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM", or a dict with the key 'term_attr'
        we can also add a key for each regex.
    span_getter : SpanGetterArg
        Which entities should be classified. By default, `doc.ents`
    on_ents_only : Union[bool, str, List[str], Set[str]]
        Whether to look for matches around detected entities only.
        Useful for faster inference in downstream tasks.

        - If True, will look in all ents located in `doc.ents` only
        - If an iterable of string is passed, will additionally look in `doc.spans[key]`
        for each key in the iterable
    explain : bool
        Whether to keep track of cues for each entity.
    **terms : Dict[str, Optional[List[str]]]
        Terms to look for.
    """

    defaults = dict()

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = None,
        *,
        attr: str,
        span_getter: SpanGetterArg,
        on_ents_only: Union[bool, str, List[str], Set[str]],
        explain: bool,
        terms: Dict[str, Optional[List[str]]],
        regex: Dict[str, Optional[List[str]]] = {},
    ):
        super().__init__(nlp=nlp, name=name)

        if attr.upper() == "NORM":
            check_normalizer(nlp)

        self.phrase_matcher = EDSPhraseMatcher(vocab=nlp.vocab, attr=attr)
        self.phrase_matcher.build_patterns(nlp=nlp, terms=terms)

        self.regex_matcher = RegexMatcher(attr=attr)
        self.regex_matcher.build_patterns(regex=regex)

        self.on_ents_only = on_ents_only

        if span_getter is None and on_ents_only is None:
            on_ents_only = True

        if on_ents_only:
            assert span_getter is None or on_ents_only is True, (
                "Cannot use both `span_getter` and `on_ents_only` as a span selection "
                "argument."
            )
            assert isinstance(on_ents_only, (list, str, set, bool)), (
                "The `on_ents_only` argument should be a "
                "string, a bool, a list or a set of string"
            )
            span_getter = "ents" if on_ents_only is True else on_ents_only
        else:
            span_getter = "ents"
        self.span_getter = validate_span_getter(span_getter)
        self.explain = explain

    def get_matches(self, doc: Doc) -> List[Span]:
        """
        Extract matches.

        Parameters
        ----------
        doc : Doc
            spaCy `Doc` object.
        Returns
        -------
        List[Span]
            List of detected spans
        """

        if self.on_ents_only:
            sents = set([ent.sent for ent in self.get_spans(doc)])

            match_iterator = (
                (
                    *self.phrase_matcher(s, as_spans=True),
                    *self.regex_matcher(s, as_spans=True),
                )
                for s in sents
            )

            matches = chain.from_iterable(match_iterator)

        else:
            matches = (
                *self.phrase_matcher(doc, as_spans=True),
                *self.regex_matcher(doc, as_spans=True),
            )

        return list(matches)

    def __call__(self, doc: Doc) -> Doc:
        return self.process(doc)



import spacy
from pytest import fixture

from edsnlp.pipelines.core.endlines.functional import build_path
from edsnlp.pipelines.core.endlines.model import EndLinesModel

texts = [
    """Le patient est arrivé hier soir.
Il est accompagné par son fils

ANTECEDENTS
Il a fait une TS en 2010;
Fumeur, il est arreté il a 5 mois
Chirurgie de coeur en 2011
CONCLUSION
Il doit prendre
le medicament indiqué 3 fois par jour. Revoir médecin
dans 1 mois.
DIAGNOSTIC :

Antecedents Familiaux:
- 1. Père avec diabete

""",
    """J'aime le \nfromage...\n""",
]


@fixture
def model_path(tmp_path, blank_nlp):
    blank_nlp = spacy.blank("eds")

    # Train model
    docs = list(blank_nlp.pipe(texts))

    # Train and predict an EndLinesModel
    endlines = EndLinesModel(nlp=blank_nlp)
    df = endlines.fit_and_predict(docs)

    assert df["PREDICTED_END_LINE"].dtype == bool

    # path_model = os.path.join(tmp_path, "endlinesmodel.pkl")
    path_model = build_path(tmp_path, "endlinesmodel.pkl")
    endlines.save(path=path_model)

    return path_model


def test_endlines(blank_nlp, model_path):
    blank_nlp = spacy.blank("eds")

    # Use an existing trained model
    blank_nlp.add_pipe("endlines", config=dict(model_path=model_path))
    docs = list(blank_nlp.pipe(texts))

    assert [i for i, t in enumerate(docs[1]) if t.tag_ == "EXCLUDED"] == [3, 8]


def test_normalizer_endlines(blank_nlp, model_path):
    blank_nlp = spacy.blank("eds")

    # Use an existing trained model
    blank_nlp.add_pipe("normalizer")
    blank_nlp.add_pipe("endlines", config=dict(model_path=model_path))
    docs = list(blank_nlp.pipe(texts))

    assert [i for i, t in enumerate(docs[1]) if t.tag_ == "EXCLUDED"] == [3, 8]



from typing import List

from pytest import fixture, mark

from edsnlp.pipelines.qualifiers.hypothesis import Hypothesis
from edsnlp.utils.examples import parse_example

examples: List[str] = [
    "Possible <ent hypothesis_=HYP>covid-19</ent>",
    (
        "Plusieurs <ent hypothesis_=HYP>diagnostics</ent> sont envisagés. "
        "Le patient est informé."
    ),
    "même si <ent hypothesis=False>le patient est jeune</ent>.",
    "Suspicion de <ent hypothesis_=HYP>diabète</ent>.",
    "Le ligament est <ent hypothesis_=CERT>rompu</ent>.",
    "Probablement du diabète mais pas de <ent hypothesis_=CERT>cécité</ent>.",
]


@fixture
def hypothesis_factory(blank_nlp):

    default_config = dict(
        pseudo=None,
        preceding=None,
        following=None,
        termination=None,
        verbs_hyp=None,
        verbs_eds=None,
        attr="NORM",
        within_ents=False,
        explain=True,
    )

    def factory(on_ents_only, **kwargs):

        config = dict(**default_config)
        config.update(kwargs)

        return Hypothesis(
            nlp=blank_nlp,
            on_ents_only=on_ents_only,
            **config,
        )

    return factory


@mark.parametrize("on_ents_only", [True, False])
def test_hypothesis(blank_nlp, hypothesis_factory, on_ents_only):

    hypothesis = hypothesis_factory(on_ents_only)

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = hypothesis(doc)

        for entity, ent in zip(entities, doc.ents):

            for modifier in entity.modifiers:

                assert bool(ent._.hypothesis_cues) == (modifier.value in {"HYP", True})

                assert getattr(ent._, modifier.key) == modifier.value

                if not on_ents_only:
                    for token in ent:
                        assert (
                            getattr(token._, modifier.key) == modifier.value
                        ), f"{modifier.key} labels don't match."


def test_hypothesis_within_ents(blank_nlp, hypothesis_factory):

    hypothesis = hypothesis_factory(True, within_ents=True)

    examples = [
        "<ent hypothesis_=HYP>Diabète, probablement de type 2</ent>.",
    ]

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = hypothesis(doc)

        for entity, ent in zip(entities, doc.ents):

            for modifier in entity.modifiers:

                assert bool(ent._.hypothesis_cues) == (modifier.value in {"HYP", True})



main_pattern = dict(
    source="main",
    regex=[
        r"lymphom(?:.{1,10}hodgkin)",
        r"lymphom",
        r"lymphangio",
        r"sezary",
        r"burkitt",
        r"kaposi",
        r"hodgkin",
        r"amylose",
        r"plasm[ao]cytome",
        r"lympho.{1,3}sarcome",
        r"lympho.?prolif",
        r"hemopathie.{1,10}lymphoide",
        r"macroglobulinemie",
        r"immunocytome",
        r"maladie.des.chaine",
        r"histiocytose.{1,5}(maligne|langerhans)",
        r"waldenst(ro|or)m",
        r"mycos.{1,10}fongoide",
        r"myelome",
        r"maladie.{1,5}immunoproliferative.{1,5}maligne",
        r"leucemie.{1,10}plasmocyte",
    ],
    regex_attr="NORM",
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bLNH\b",
        r"\bLH\b",
        r"\bEATL\b",
        r"\bLAGC\b",
        r"\bLDGCB\b",
    ],
    regex_attr="TEXT",
    exclude=dict(
        regex=["/L", "/mL"],
        window=10,
    ),
)


gammapathy = dict(
    source="gammapathy",
    regex=[
        r"gammapathie monoclonale",
    ],
    exclude=dict(
        regex=[
            "benin",
            "benign",
            "signification.indeter",
            "NMSI",
            "MGUS",
        ],
        window=(0, 5),
    ),
    regex_attr="NORM",
)


default_patterns = [
    main_pattern,
    acronym,
    # gammapathy,
]
